import fnmatch


def _normalize_name(name):
    """
    Normalizes action names to handle hyphens and underscores interchangeably.
    """
    return name.replace("-", "_").strip()


def accept_action(pattern: str, package_name: str, action_name: str) -> bool:
    """
    Determines whether an action should be accepted based on the whitelist pattern.

    Args:
      pattern: The whitelist pattern to match against.
      package_name: The name of the package containing the action (can be empty).
      action_name: The name of the action.

    Returns:
      True if the action matches the whitelist pattern, False otherwise.
    """
    package_name = _normalize_name(package_name)
    action_name = _normalize_name(action_name)
    pattern = _normalize_name(pattern)

    for single_pattern in pattern.split(","):
        # Check if the pattern includes a package name separated by '/'
        if "/" in single_pattern:
            package_pattern, action_pattern = single_pattern.split("/", 1)
        else:
            package_pattern = "*"
            action_pattern = single_pattern

        # Check if the package name matches the package pattern (if provided)
        if fnmatch.fnmatch(package_name, package_pattern):
            # Match the action name against the action pattern
            if fnmatch.fnmatch(action_name, action_pattern):
                return True

    return False


def accept_action_package(pattern: str, package_name: str) -> bool:
    """
    Determines whether an action package should be accepted based on the whitelist pattern.

    Args:
      pattern: The whitelist pattern to match against.
      package_name: The name of the package containing the action (can be empty).
      action_name: The name of the action.

    Returns:
      True if the action matches the whitelist pattern, False otherwise.
    """
    package_name = _normalize_name(package_name)
    pattern = _normalize_name(pattern)

    for single_pattern in pattern.split(","):
        # Check if the pattern includes a package name separated by '/'
        if "/" in single_pattern:
            package_pattern, _action_pattern = single_pattern.split("/", 1)
        else:
            package_pattern = "*"

        # Check if the package name matches the package pattern (if provided)
        if fnmatch.fnmatch(package_name, package_pattern):
            return True

    return False



import asyncio
import codecs
import json
import logging
import os
import socket
import sys
import typing
from typing import Optional

import requests
from pydantic import BaseModel, ValidationError
from termcolor import colored

if typing.TYPE_CHECKING:
    import websockets

log = logging.getLogger(__name__)


class SessionPayload(BaseModel):
    sessionId: str
    sessionSecret: str


class BodyPayload(BaseModel):
    requestId: str
    path: str
    method: str = "GET"
    body: dict | None = None
    headers: dict


class ExposeSessionJson(BaseModel):
    expose_session: str
    url: str


def get_expose_session_path(datadir: str) -> str:
    return os.path.join(datadir, "expose_session.json")


def read_expose_session_json(datadir: str) -> None | ExposeSessionJson:
    session_json = None
    try:
        expose_session_path = get_expose_session_path(datadir)
        log.debug(f"🗂️ Reading expose_session.json path={expose_session_path}")
        with open(expose_session_path, "r") as f:
            session_json = ExposeSessionJson(**json.load(f))
    except FileNotFoundError:
        pass
    except ValidationError as e:
        log.error("Failed to load previous expose session", e)
    except json.JSONDecodeError as e:
        log.error("Failed to decode exopse session json", e)
    return session_json


def write_expose_session_json(
    datadir: str, expose_session_json: ExposeSessionJson
) -> None:
    expose_session_path = get_expose_session_path(datadir)
    log.debug(f"🗂️ Writing expose_session.json path={expose_session_path}")
    with open(expose_session_path, "w") as f:
        json.dump(
            expose_session_json.model_dump(),
            f,
            indent=2,
        )


def get_expose_session(payload: SessionPayload) -> str:
    return f"{payload.sessionId}:{payload.sessionSecret}".encode("ascii").hex()


def get_expose_session_payload(expose_session: str) -> SessionPayload:
    session_id, session_secret = codecs.decode(
        expose_session.encode("ascii"), "hex"
    ).split(b":")
    return SessionPayload(
        sessionId=session_id.decode(), sessionSecret=session_secret.decode()
    )


def forward_request(base_url: str, payload: BodyPayload) -> requests.Response:
    url = base_url.rstrip("/") + "/" + payload.path.lstrip("/")

    if payload.method == "GET":
        return requests.get(url, headers=payload.headers)
    elif payload.method == "POST":
        return requests.post(url, json=payload.body, headers=payload.headers)
    elif payload.method == "PUT":
        return requests.put(url, json=payload.body, headers=payload.headers)
    elif payload.method == "DELETE":
        return requests.delete(url, headers=payload.headers)
    else:
        raise NotImplementedError(f"Method {payload.method} not implemented")


async def handle_ping_pong(
    ws: "websockets.WebSocketClientProtocol",
    pong_queue: asyncio.Queue,
    ping_interval: int,
):
    while True:
        await asyncio.sleep(ping_interval)
        await ws.send("ping")

        try:
            await asyncio.wait_for(pong_queue.get(), timeout=ping_interval)
        except (asyncio.TimeoutError, TimeoutError):
            log.debug("Lost connection to expose server")
            await ws.close()
            break


def handle_session_payload(
    session_payload: SessionPayload, expose_url: str, datadir: str, api_key: str | None
):
    url = f"https://{session_payload.sessionId}.{expose_url}"
    log.info(
        colored("  🌍 Public URL: ", "green", attrs=["bold"])
        + colored(f"{url}", "light_blue")
    )
    if api_key:
        log.info(
            colored("  🔑 API Authorization Bearer key: ", attrs=["bold"])
            + f"{api_key}\n"
        )
    new_expose_session = get_expose_session(session_payload)
    write_expose_session_json(
        datadir=datadir,
        expose_session_json=ExposeSessionJson(
            expose_session=new_expose_session,
            url=url,
        ),
    )


def handle_body_payload(
    payload: BodyPayload,
    base_url: str,
):
    response: requests.Response = forward_request(base_url=base_url, payload=payload)
    return json.dumps(
        {
            "requestId": payload.requestId,
            "response": json.dumps(
                response.json(),
                indent=2,
            ),
            "status": response.status_code,
        }
    )


def _headers_from_session_payload(session_payload: SessionPayload):
    headers = {
        "x-session-id": session_payload.sessionId,
        "x-session-secret": session_payload.sessionSecret,
    }
    return headers


async def expose_server(
    port: int,
    host: str,
    expose_url: str,
    datadir: str,
    expose_session: str | None = None,
    ping_interval: int = 4,
    api_key: str | None = None,
):
    """
    Exposes the server to the world.
    """
    import websockets

    pong_queue: asyncio.Queue = asyncio.Queue(maxsize=1)

    async def listen_for_requests() -> None:
        session_payload: Optional[SessionPayload] = (
            get_expose_session_payload(expose_session) if expose_session else None
        )

        headers = (
            _headers_from_session_payload(session_payload) if session_payload else {}
        )

        # In testing wss:// is added (in production it's just robocorp.link and
        # we connect to wss://client.robocorp.link and when we receive the session
        # id/secret we print it as: https://{session_payload.sessionId}.{expose_url}).
        if not expose_url.startswith("wss://"):
            use_url = f"wss://client.{expose_url}"
        else:
            use_url = expose_url

        async for ws in websockets.connect(
            use_url,
            extra_headers=headers,
            logger=log,
            open_timeout=2,
            close_timeout=0,
        ):
            ping_task = asyncio.create_task(
                handle_ping_pong(ws, pong_queue, ping_interval)
            )

            try:
                while True:
                    message = await ws.recv()
                    if message == "pong":
                        await pong_queue.put("pong")
                        continue

                    data = json.loads(message)

                    match data:
                        case {"sessionId": _, "sessionSecret": _}:
                            try:
                                session_payload = SessionPayload(**data)
                                # Make sure that after we connect the session is
                                # always the same for reconnects.
                                # Note: tricky detail: because we're reconnecting
                                # in the above `for` which already has the `extra_headers`
                                # bound, we just update the headers in-place...
                                # this could fail in the future is websockets.connect
                                # for some reason does a copy of the extra headers,
                                # so, we have to be careful (probably not the best
                                # approach, but working for a fast fix).
                                headers.update(
                                    _headers_from_session_payload(session_payload)
                                )
                                handle_session_payload(
                                    session_payload, expose_url, datadir, api_key
                                )
                            except ValidationError as e:
                                if not session_payload:
                                    log.error("Expose session initialization failed", e)
                                    continue
                        case _:
                            try:
                                body_payload = BodyPayload(**data)
                                response = handle_body_payload(
                                    body_payload, f"http://{host}:{port}"
                                )
                                await ws.send(response)
                            except ValidationError as e:
                                log.error("Expose request validation failed", e)
                                continue
            except websockets.exceptions.ConnectionClosedError:
                log.info("Lost connection. Reconnecting to expose server...")
            except socket.gaierror:
                log.info("No internet connection")
                break
            finally:
                ping_task.cancel()
                try:
                    await ping_task
                except asyncio.CancelledError:
                    pass
        else:
            log.info("Expose server connection closed")

    task = asyncio.create_task(listen_for_requests())
    await task  # Wait for listen_for_requests to complete


def main(
    parent_pid: str,
    port: str,
    verbose: str,
    host: str,
    expose_url: str,
    datadir: str,
    expose_session: str,
    api_key: str,
    config_logging: bool = True,
) -> None:
    """
    Args:
        parent_pid:
            The pid of the parent (when the given pid exists, this process must
            also exit itself). If it's an empty string, it's ignored.

        verbose:
            A string with `v` chars (currently a single `v` means debug, otherwise
            info).

        host:
            The host to where the data should be forwarded.

        port:
            The port to where the data should be forwarded.

        expose_url:
            The url for the expose (usually "robocorp.link").

            Note that the protocol here is that it'll connect to something as:

            wss://client.{expose_url}

            Which will then provide a message with a payload such as:

                {"sessionId": _, "sessionSecret": _}

            and then the server will start accepting connections at:

            https://{session_payload.sessionId}.{expose_url}

            and then messages can be sent to that address to be received
            by the action server.

        datadir:
            The directory where the expose session information should
            be stored.

        expose_session:
            A string with the information of a previous session which
            should be restored (or 'None' if there's no previous session
            information to be restored).

        api_key:
            The api key that should be passed to the action server along with
            the forwarded messages so that the action server accepts it.
    """
    from robocorp.action_server._robo_utils.process import exit_when_pid_exists

    if config_logging:
        logging.basicConfig(
            level=logging.DEBUG if verbose.count("v") > 0 else logging.INFO,
            format="%(message)s",
            datefmt="[%X]",
        )

        def remove_traceback(record):
            if record.levelname == "INFO":
                if record.exc_info:
                    record.exc_info = None
            return True

        # Don't bee too verbose when attempting to reconnect.
        log.addFilter(remove_traceback)

    if parent_pid:
        exit_when_pid_exists(int(parent_pid))

    try:
        asyncio.run(
            expose_server(
                port=int(port),
                host=host,
                expose_url=expose_url,
                datadir=datadir,
                expose_session=expose_session if expose_session != "None" else None,
                api_key=api_key,
            )
        )
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    try:
        (
            parent_pid,
            port,
            verbose,
            host,
            expose_url,
            datadir,
            expose_session,
            api_key,
        ) = sys.argv[1:]
    except Exception:
        raise RuntimeError(f"Unable to initialize with sys.argv: {sys.argv}")

    main(
        parent_pid,
        port,
        verbose,
        host,
        expose_url,
        datadir,
        expose_session,
        api_key,
    )



import logging
import os
import sys
import typing
from pathlib import Path
from typing import List, Optional, Tuple

from .cli_errors import ActionPackageError

log = logging.getLogger(__name__)


def _verify_name(package_yaml: Optional[Path], name: str) -> None:
    # Could be used to verify that something is not declared (no-op for now).
    pass


_version_chars_without_eq = (">", "<", "^", "~", "!", ",", ";")


def _interpret_entry(package_yaml: Optional[Path], entry: str) -> Tuple[str, str, str]:
    # https://packaging.python.org/en/latest/specifications/version-specifiers/#version-specifiers
    # https://packaging.python.org/en/latest/specifications/version-specifiers/#compatible-release
    for c in _version_chars_without_eq:
        if c in entry:
            if package_yaml is None:
                raise ActionPackageError(
                    f"Error in entry: {entry}. Char: {c!r} not valid, currently only '=' is a valid operator."
                )
            else:
                raise ActionPackageError(
                    f"Error in entry: {entry}. Char: {c!r} not valid, currently only '=' is a valid operator. (in {package_yaml})."
                )

    i = entry.find("=")
    if i == -1:
        # No version specified.
        name = entry
        _verify_name(package_yaml, name)
        return name, "", ""

    name, version = entry[:i], entry[i + 1 :]
    if "=" in version:
        raise ActionPackageError(
            f"Error in entry: {entry}. Only a single '=' should be used (in {package_yaml})."
        )
    _verify_name(package_yaml, name)
    return name, "=", version


def convert_conda_entry(
    package_yaml: Optional[Path], entry: str
) -> Tuple[str, str, str, str]:
    name, op, version = _interpret_entry(package_yaml, entry)

    if op:
        if op == "=":
            return name, f"{name}={version}", "=", version
        if package_yaml is None:
            raise ActionPackageError(f"Error in entry: {entry}. {op!r} not supported.")
        else:
            raise ActionPackageError(
                f"Error in entry: {entry}. {op!r} not supported (in {package_yaml})."
            )

    assert not version
    return name, name, "", ""


def convert_pip_entry(
    package_yaml: Optional[Path], entry: str
) -> Tuple[str, str, str, str]:
    name, op, version = _interpret_entry(package_yaml, entry)
    if op:
        if op == "=":
            return name, f"{name}=={version}", "==", version
        if package_yaml is None:
            raise ActionPackageError(f"Error in entry: {entry}. {op!r} not supported.")
        else:
            raise ActionPackageError(
                f"Error in entry: {entry}. {op!r} not supported (in {package_yaml})."
            )

    assert not version
    return name, name, "", ""


def _create_package_from_conda_yaml(
    conda_yaml: Path, dry_run: bool, backup: bool, *, stream: Optional[typing.IO] = None
):
    import yaml

    from ..package_deps.conda_impl import conda_match_spec
    from ..package_deps.pip_impl.pip_distlib_util import parse_requirement
    from ..termcolors import bold_yellow

    with conda_yaml.open("r", encoding="utf-8") as s:
        contents = yaml.safe_load(s)

    dependencies = contents.get("dependencies")
    pip_deps: List[str] = []
    conda_deps: List[str] = []
    if isinstance(dependencies, list):
        for dep in dependencies:
            if isinstance(dep, dict):
                if tuple(dep.keys()) != ("pip",):
                    raise ActionPackageError(
                        f'Unable to convert to package.yaml: Keys of dependency dict is not just "pip" dependency. Found: {dep}.'
                    )

                for pip_dep in dep["pip"]:
                    if not isinstance(pip_dep, str):
                        raise ActionPackageError(
                            f"Unable to convert to package.yaml: Expected pip dep to be str. Found: {pip_dep} ({type(pip_dep)})."
                        )
                    pip_deps.append(pip_dep)

                continue

            if not isinstance(dep, str):
                raise ActionPackageError(
                    f"Unable to convert to package.yaml: Expected dep to be str. Found: {dep} ({type(dep)})."
                )

            conda_deps.append(dep)

    # Ok, at this point we should have the deps. Let's put them in the new format.
    new_pypi_deps = []
    for dep in pip_deps:
        try:
            if dep.startswith("--"):
                continue
            parsed = parse_requirement(dep)
            if parsed is None:
                continue

            for op, version in parsed.constraints:
                if op != "==":
                    print(
                        bold_yellow(
                            f"Found constraint: {op!r} for {parsed.name}. It'll be converted to '='"
                        ),
                        file=stream,
                    )
                new_pypi_deps.append(f"{parsed.name}={version}")
                break

            for op, version in parsed.constraints[1:]:
                print(
                    bold_yellow(f"Ignoring constraint: {op!r} for {parsed.name}."),
                    file=stream,
                )

        except Exception:
            raise ActionPackageError(f"Unable to parse pip dep: {dep}.")

    new_conda_deps = []
    all_version_chars = _version_chars_without_eq + ("=",)
    for dep in conda_deps:
        try:
            spec = conda_match_spec.parse_spec_str(dep)

            # It may not have a version if it wasn't specified.
            version_spec = spec.get("version", "*").strip()
            name = spec["name"]

            while version_spec.endswith(("*", ".")):
                version_spec = version_spec[:-1]

            removed = []
            while version_spec.startswith(all_version_chars):
                removed.append(version_spec[0])
                version_spec = version_spec[1:]

            if removed:
                op = "".join(removed)
                if set(op) != set("="):
                    print(
                        bold_yellow(
                            f"Found constraint: {op!r} for {name}. It'll be converted to '='"
                        ),
                        file=stream,
                    )

            if not version_spec:
                new_conda_deps.append(name)
            else:
                new_conda_deps.append(f"{name}={version_spec}")
        except Exception:
            log.exception(f"Error collecting version info from: {dep}")

    new_dependencies = {
        "dependencies": {"conda-forge": new_conda_deps, "pypi": new_pypi_deps}
    }

    post_install = contents.get("rccPostInstall")
    post_install_str = ""
    if post_install:
        post_install_str = yaml.dump({"post-install": post_install})

    output = f"""
# Required: A short name for the action package
name: Action package name
    
# Required: A description of what's in the action package.
description: Action package description

# Required: The current version of this action package.
version: 0.0.1

# Required: A link to where the documentation on the package lives.
documentation: https://github.com/...

{yaml.safe_dump(new_dependencies, sort_keys=False)}{post_install_str}
"""

    target = conda_yaml.parent / "package.yaml"

    print(f"Creating {target.name}", file=stream)
    if dry_run:
        print(f"Contents:\n----------------\n{output}----------------\n", file=stream)
    else:
        target.write_text(output, encoding="utf-8")

    # Remove conda.yaml
    if backup:
        new_conda_yaml_path = conda_yaml.parent / (conda_yaml.name + ".bak")
        print(f"Renaming {conda_yaml.name} to {new_conda_yaml_path.name}", file=stream)
        if not dry_run:
            conda_yaml.rename(new_conda_yaml_path)
    else:
        print(f"Removing {conda_yaml.name}", file=stream)
        if not dry_run:
            os.remove(conda_yaml)

    # Remove robot.yaml
    robot_yaml = conda_yaml.parent / "robot.yaml"
    if robot_yaml.exists():
        if backup:
            new_robot_yaml_path = robot_yaml.parent / (robot_yaml.name + ".bak")
            print(
                f"Renaming {robot_yaml.name} to {new_robot_yaml_path.name}", file=stream
            )
            if not dry_run:
                robot_yaml.rename(new_robot_yaml_path)
        else:
            print(f"Removing {robot_yaml.name}", file=stream)
            if not dry_run:
                os.remove(robot_yaml)


def update_package(
    cwd: Path, dry_run: bool, backup: bool = True, *, stream=None
) -> None:
    """
    Updates a given package to conform to the latest structure.

    Current operations done:
        Convert conda.yaml or action-server.yaml to package.yaml

    Args:
        cwd: This is the directory where the action package is
        (i.e.: the directory with package.yaml or some older version
        having conda.yaml or action-server.yaml).
    """
    if stream is None:
        stream = sys.stdout
    conda_yaml = cwd / "conda.yaml"
    if not conda_yaml.exists():
        conda_yaml = cwd / "action-server.yaml"

    if conda_yaml.exists():
        print(f"Generating package.yaml from {conda_yaml.name}", file=stream)
        _create_package_from_conda_yaml(
            conda_yaml, dry_run=dry_run, backup=backup, stream=stream
        )


def create_hash(contents: str) -> str:
    import hashlib

    sha256_hash = hashlib.sha256()
    sha256_hash.update(contents.encode("utf-8"))
    return sha256_hash.hexdigest()


def create_conda_contents_from_package_yaml_contents(
    package_yaml: Path, package_yaml_contents: dict
) -> dict:
    from ..termcolors import bold_yellow

    def _get_in_dict(
        dct: dict,
        entry: str,
        expected_result_type: Optional[type],
        required: bool = True,
    ):
        v = dct.get(entry)
        if v is None:
            if not required:
                return None

            raise ActionPackageError(
                f"'Required session: {entry}' not specified (in {package_yaml})."
            )
        if expected_result_type is not None and not isinstance(v, expected_result_type):
            raise ActionPackageError(
                f"{entry} is expected to be {expected_result_type}. Found: {type(v)} (in {package_yaml})."
            )

        return v

    found_names = set()

    def _validate_name(name):
        if name in found_names:
            raise ActionPackageError(
                f"Dependency: {name} specified more than once (in {package_yaml})."
            )
        found_names.add(name)

    if not isinstance(package_yaml_contents, dict):
        raise ActionPackageError(
            f"Dict not found as top-level element (in {package_yaml})."
        )

    python_deps = _get_in_dict(package_yaml_contents, "dependencies", dict)
    conda_forge = _get_in_dict(python_deps, "conda-forge", list)
    local_wheels = _get_in_dict(python_deps, "local-wheels", list, required=False)
    post_install = _get_in_dict(
        package_yaml_contents, "post-install", list, required=False
    )

    # This is an error
    pip = _get_in_dict(python_deps, "pip", list, required=False)
    if pip:
        raise ActionPackageError(
            f"'pip' entry not found in 'dependencies.pip' should be renamed to 'pypi' (in {package_yaml})."
        )

    # The user could get everything from 'conda'
    pypi = _get_in_dict(python_deps, "pypi", list, required=False) or []

    converted_conda_entries: list = []
    found_truststore = False
    for entry in conda_forge:
        name, entry, op, version = convert_conda_entry(package_yaml, entry)
        _validate_name(name)
        if name in ("truststore", "robocorp-truststore"):
            found_truststore = True
        converted_conda_entries.append(entry)

    if "python" not in found_names:
        raise ActionPackageError(
            f"'python' entry not found in 'dependencies.conda-forge' (in {package_yaml})."
        )

    converted_pip_entries: list = []
    for entry in pypi:
        if entry.replace(" ", "") == "--use-feature=truststore":
            log.info(
                bold_yellow(
                    "--use-feature=truststore flag does not need to "
                    "be specified (it is automatically used when a "
                    '"robocorp-trustore" or "trustore" dependency is added).'
                )
            )
            found_truststore = True
            continue

        if entry.startswith("--"):
            raise ActionPackageError(f"Unexpected entry in pypi: {entry}")

        name, entry, op, version = convert_pip_entry(package_yaml, entry)
        _validate_name(name)
        if name in ("truststore", "robocorp-truststore"):
            found_truststore = True

        converted_pip_entries.append(entry)

    if found_truststore:
        converted_pip_entries.insert(0, "--use-feature=truststore")

    cwd_dir = package_yaml.parent.absolute()
    if local_wheels:
        for wheel in local_wheels:
            original = wheel
            if not os.path.isabs(wheel):
                wheel = os.path.join(str(cwd_dir / wheel))

            if not os.path.exists(wheel):
                raise ActionPackageError(
                    f"'local-wheel': {original} could not be found (in {package_yaml})."
                )

            converted_pip_entries.append(Path(wheel).as_posix())

    converted_conda_entries.append({"pip": converted_pip_entries})

    data: dict = {"channels": ["conda-forge"], "dependencies": converted_conda_entries}
    if post_install:
        data["rccPostInstall"] = post_install

    return data


def create_conda_from_package_yaml(datadir: Path, package_yaml: Path) -> Path:
    """
    Args:
        package_yaml: This is the package.yaml from which the conda.yaml
            (to be supplied to rcc to create the env) should be created.

        package_yaml_contents: If specified this are the yaml-loaded contents
            of the package yaml.

    Returns: The path to the generated conda.yaml.
    """
    import yaml

    if not package_yaml.exists():
        raise ActionPackageError(f"File does not exist ({package_yaml}).")

    try:
        with open(package_yaml, "r", encoding="utf-8") as stream:
            package_yaml_contents = yaml.safe_load(stream)
    except Exception:
        raise ActionPackageError(f"Error loading file as yaml ({package_yaml}).")

    data = create_conda_contents_from_package_yaml_contents(
        package_yaml, package_yaml_contents
    )

    tmpdir = datadir / "tmpdir"
    if not tmpdir.exists():
        tmpdir.mkdir(exist_ok=True)

    new_package_yaml_contents = yaml.dump(data)

    conda_path = tmpdir / f"conda_{create_hash(new_package_yaml_contents)[:12]}.yaml"
    conda_path.write_text(new_package_yaml_contents)

    return conda_path



import re
from types import SimpleNamespace as Container
from urllib.parse import urlparse

#
# Requirement parsing code as per PEP 508
#


IDENTIFIER = re.compile(r"^([\w\.-]+)\s*")
VERSION_IDENTIFIER = re.compile(r"^([\w\.*+-]+)\s*")
COMPARE_OP = re.compile(r"^(<=?|>=?|={2,3}|[~!]=)\s*")
MARKER_OP = re.compile(r"^((<=?)|(>=?)|={2,3}|[~!]=|in|not\s+in)\s*")
OR = re.compile(r"^or\b\s*")
AND = re.compile(r"^and\b\s*")
NON_SPACE = re.compile(r"(\S+)\s*")
STRING_CHUNK = re.compile(r"([\s\w\.{}()*+#:;,/?!~`@$%^&=|<>\[\]-]+)")


def parse_marker(marker_string):
    """
    Parse a marker string and return a dictionary containing a marker expression.

    The dictionary will contain keys "op", "lhs" and "rhs" for non-terminals in
    the expression grammar, or strings. A string contained in quotes is to be
    interpreted as a literal string, and a string not contained in quotes is a
    variable (such as os_name).
    """

    def marker_var(remaining):
        # either identifier, or literal string
        m = IDENTIFIER.match(remaining)
        if m:
            result = m.groups()[0]
            remaining = remaining[m.end() :]
        elif not remaining:
            raise SyntaxError("unexpected end of input")
        else:
            q = remaining[0]
            if q not in "'\"":
                raise SyntaxError("invalid expression: %s" % remaining)
            oq = "'\"".replace(q, "")
            remaining = remaining[1:]
            parts = [q]
            while remaining:
                # either a string chunk, or oq, or q to terminate
                if remaining[0] == q:
                    break
                elif remaining[0] == oq:
                    parts.append(oq)
                    remaining = remaining[1:]
                else:
                    m = STRING_CHUNK.match(remaining)
                    if not m:
                        raise SyntaxError("error in string literal: %s" % remaining)
                    parts.append(m.groups()[0])
                    remaining = remaining[m.end() :]
            else:
                s = "".join(parts)
                raise SyntaxError("unterminated string: %s" % s)
            parts.append(q)
            result = "".join(parts)
            remaining = remaining[1:].lstrip()  # skip past closing quote
        return result, remaining

    def marker_expr(remaining):
        if remaining and remaining[0] == "(":
            result, remaining = marker(remaining[1:].lstrip())
            if remaining[0] != ")":
                raise SyntaxError("unterminated parenthesis: %s" % remaining)
            remaining = remaining[1:].lstrip()
        else:
            lhs, remaining = marker_var(remaining)
            while remaining:
                m = MARKER_OP.match(remaining)
                if not m:
                    break
                op = m.groups()[0]
                remaining = remaining[m.end() :]
                rhs, remaining = marker_var(remaining)
                lhs = {"op": op, "lhs": lhs, "rhs": rhs}
            result = lhs
        return result, remaining

    def marker_and(remaining):
        lhs, remaining = marker_expr(remaining)
        while remaining:
            m = AND.match(remaining)
            if not m:
                break
            remaining = remaining[m.end() :]
            rhs, remaining = marker_expr(remaining)
            lhs = {"op": "and", "lhs": lhs, "rhs": rhs}
        return lhs, remaining

    def marker(remaining):
        lhs, remaining = marker_and(remaining)
        while remaining:
            m = OR.match(remaining)
            if not m:
                break
            remaining = remaining[m.end() :]
            rhs, remaining = marker_and(remaining)
            lhs = {"op": "or", "lhs": lhs, "rhs": rhs}
        return lhs, remaining

    return marker(marker_string)


def parse_requirement(req):
    """
    Parse a requirement passed in as a string. Return a Container
    whose attributes contain the various parts of the requirement.
    """
    remaining = req.strip()
    if not remaining or remaining.startswith("#"):
        return None
    m = IDENTIFIER.match(remaining)
    if not m:
        raise SyntaxError("name expected: %s" % remaining)
    distname = m.groups()[0]
    remaining = remaining[m.end() :]
    extras = mark_expr = versions = uri = None
    if remaining and remaining[0] == "[":
        i = remaining.find("]", 1)
        if i < 0:
            raise SyntaxError("unterminated extra: %s" % remaining)
        s = remaining[1:i]
        remaining = remaining[i + 1 :].lstrip()
        extras = []
        while s:
            m = IDENTIFIER.match(s)
            if not m:
                raise SyntaxError("malformed extra: %s" % s)
            extras.append(m.groups()[0])
            s = s[m.end() :]
            if not s:
                break
            if s[0] != ",":
                raise SyntaxError("comma expected in extras: %s" % s)
            s = s[1:].lstrip()
        if not extras:
            extras = None
    if remaining:
        if remaining[0] == "@":
            # it's a URI
            remaining = remaining[1:].lstrip()
            m = NON_SPACE.match(remaining)
            if not m:
                raise SyntaxError("invalid URI: %s" % remaining)
            uri = m.groups()[0]
            t = urlparse(uri)
            # there are issues with Python and URL parsing, so this test
            # is a bit crude. See bpo-20271, bpo-23505. Python doesn't
            # always parse invalid URLs correctly - it should raise
            # exceptions for malformed URLs
            if not (t.scheme and t.netloc):
                raise SyntaxError("Invalid URL: %s" % uri)
            remaining = remaining[m.end() :].lstrip()
        else:

            def get_versions(ver_remaining):
                """
                Return a list of operator, version tuples if any are
                specified, else None.
                """
                m = COMPARE_OP.match(ver_remaining)
                versions = None
                if m:
                    versions = []
                    while True:
                        op = m.groups()[0]
                        ver_remaining = ver_remaining[m.end() :]
                        m = VERSION_IDENTIFIER.match(ver_remaining)
                        if not m:
                            raise SyntaxError("invalid version: %s" % ver_remaining)
                        v = m.groups()[0]
                        versions.append((op, v))
                        ver_remaining = ver_remaining[m.end() :]
                        if not ver_remaining or ver_remaining[0] != ",":
                            break
                        ver_remaining = ver_remaining[1:].lstrip()
                        # Some packages have a trailing comma which would break things
                        # See issue #148
                        if not ver_remaining:
                            break
                        m = COMPARE_OP.match(ver_remaining)
                        if not m:
                            raise SyntaxError("invalid constraint: %s" % ver_remaining)
                    if not versions:
                        versions = None
                return versions, ver_remaining

            if remaining[0] != "(":
                versions, remaining = get_versions(remaining)
            else:
                i = remaining.find(")", 1)
                if i < 0:
                    raise SyntaxError("unterminated parenthesis: %s" % remaining)
                s = remaining[1:i]
                remaining = remaining[i + 1 :].lstrip()
                # As a special diversion from PEP 508, allow a version number
                # a.b.c in parentheses as a synonym for ~= a.b.c (because this
                # is allowed in earlier PEPs)
                if COMPARE_OP.match(s):
                    versions, _ = get_versions(s)
                else:
                    m = VERSION_IDENTIFIER.match(s)
                    if not m:
                        raise SyntaxError("invalid constraint: %s" % s)
                    v = m.groups()[0]
                    s = s[m.end() :].lstrip()
                    if s:
                        raise SyntaxError("invalid constraint: %s" % s)
                    versions = [("~=", v)]

    if remaining:
        if remaining[0] != ";":
            raise SyntaxError("invalid requirement: %s" % remaining)
        remaining = remaining[1:].lstrip()

        mark_expr, remaining = parse_marker(remaining)

    if remaining and remaining[0] != "#":
        raise SyntaxError("unexpected trailing data: %s" % remaining)

    if not versions:
        rs = distname
    else:
        rs = "%s %s" % (distname, ", ".join(["%s %s" % con for con in versions]))
    return Container(
        name=distname,
        extras=extras,
        constraints=versions,
        marker=mark_expr,
        url=uri,
        requirement=rs,
    )



from ._conda_deps import CondaDeps
from ._pip_deps import PipDeps


class PackageDepsConda(CondaDeps):
    def add_dep(self, value: str, as_range):
        from ..action_package_handling import convert_conda_entry
        from ._conda_deps import CondaDepInfo

        try:
            _name, entry_in_conda, _op, _version = convert_conda_entry(None, value)
        except Exception as e:
            self._deps[value] = CondaDepInfo(
                value, value, "", as_range, error_msg=f"Error parsing '{value}': {e}"
            )

        else:
            CondaDeps.add_dep(self, entry_in_conda, as_range)


class PackageDepsPip(PipDeps):
    """
    Handles the format of versions supported by package.yaml.
    """

    def add_dep(self, value: str, as_range):
        from ..action_package_handling import convert_pip_entry
        from ._pip_deps import PipDepInfo

        try:
            _name, entry_in_pip, _op, _version = convert_pip_entry(None, value)
        except Exception as e:
            self._deps[value] = PipDepInfo(
                name=value,
                extras=None,
                constraints=None,
                marker=None,
                url="",
                requirement=value,
                dep_range=as_range,
                error_msg=f"Error parsing '{value}': {e}",
            )

        else:
            PipDeps.add_dep(self, entry_in_pip, as_range)



import datetime
import enum
import itertools
import json
import logging
import os
import shutil
import threading
import time
import typing
from concurrent.futures import Future
from contextlib import contextmanager
from functools import partial
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Sequence, Set, Tuple, Union

from ._deps_protocols import (
    CondaVersionInfo,
    IOnFinished,
    LatestIndexInfoTypedDict,
    SubdirToBuildAndDependsJsonBytesType,
)

if typing.TYPE_CHECKING:
    # No need for sqlite3 until it's actually used.
    from sqlite3 import Cursor
else:
    Cursor = object

log = logging.getLogger(__name__)

TABLE_PACKAGES_SQL = """
    CREATE TABLE Packages (
        package_id INTEGER PRIMARY KEY,
        package_name TEXT UNIQUE
    );
"""

TABLE_VERSIONS_SQL = """
    CREATE TABLE Versions (
        version_id INTEGER PRIMARY KEY,
        package_id INTEGER,
        depends TEXT,
        timestamp INTEGER,
        version TEXT,
        subdir TEXT,
        build TEXT,
        FOREIGN KEY (package_id) REFERENCES Packages(package_id)
    );
"""

CREATE_INDEXES_SQL = [
    """
CREATE UNIQUE INDEX package_name_index 
ON Packages(package_name);
""",
    """
CREATE INDEX package_id_on_versions_index 
ON Versions(package_id);
""",
    """
CREATE INDEX version_index 
ON Versions(version);
""",
]


class State(enum.Enum):
    initial = 1
    downloading = 2
    done = 3


def timestamp_to_datetime(timestamp_milliseconds: float) -> datetime.datetime:
    timestamp_seconds = timestamp_milliseconds / 1000  # Convert to seconds
    return datetime.datetime.fromtimestamp(timestamp_seconds)


class SqliteQueries:
    def __init__(self, sqlite_file: Union[Path, Sequence[Path]]):
        sqlite_files: Sequence[Path]
        if isinstance(sqlite_file, Path):
            sqlite_files = [sqlite_file]
        else:
            sqlite_files = sqlite_file
        self.sqlite_files: Sequence[Path] = sqlite_files

    @contextmanager
    def db_cursors(
        self, db_cursor: Optional[Sequence[Cursor]] = None
    ) -> Iterator[Sequence[Cursor]]:
        if db_cursor is not None:
            yield db_cursor
            return

        import sqlite3

        db_cursors: List[Cursor] = []
        db_connections: list = []
        for path in self.sqlite_files:
            db_connections.append(sqlite3.connect(path))
            db_cursors.append(db_connections[-1].cursor())
        try:
            yield db_cursors
        finally:
            for cursor in db_cursors:
                try:
                    cursor.close()
                except Exception:
                    log.exception("Error closing sqlite cursor.")

            for db_connection in db_connections:
                try:
                    db_connection.close()
                except Exception:
                    log.exception("Error closing sqlite connection.")

    def query_names(self, db_cursors: Optional[Sequence[Cursor]] = None) -> Set[str]:
        with self.db_cursors(db_cursors) as db_cursors:
            package_names: Set[str] = set()
            for db_cursor in db_cursors:
                db_cursor.execute("SELECT package_name FROM packages")
                rows = db_cursor.fetchall()

                package_names.update(row[0] for row in rows)
            return package_names

    def query_versions(
        self, package_name, db_cursors: Optional[Sequence[Cursor]] = None
    ) -> Set[str]:
        with self.db_cursors(db_cursors) as db_cursors:
            versions: Set[str] = set()
            for db_cursor in db_cursors:
                db_cursor.execute(
                    """
SELECT Versions.version
FROM Packages
INNER JOIN Versions ON Packages.package_id = Versions.package_id
WHERE Packages.package_name = ?
""",
                    (package_name,),
                )
                rows = db_cursor.fetchall()

                # Note that we exclude release candidates and development versions.
                versions.update(
                    row[0]
                    for row in rows
                    if "_rc" not in row[0] and "_dev" not in row[0]
                )

            return versions

    def query_version_info(
        self,
        package_name: str,
        version: str,
        db_cursors: Optional[Sequence[Cursor]] = None,
    ) -> CondaVersionInfo:
        """
        Note that the same package/version may have multiple infos
        because of build ids (but in this API this isn't given,
        so, if the same depends are available then they're shown
        only as one).
        """
        with self.db_cursors(db_cursors) as db_cursors:
            subdir_to_build_and_depends: SubdirToBuildAndDependsJsonBytesType = {}
            max_timestamp = 0

            for db_cursor in db_cursors:
                db_cursor.execute(
                    """
SELECT Versions.depends, Versions.timestamp, Versions.subdir, Versions.build
FROM Packages
INNER JOIN Versions ON Packages.package_id = Versions.package_id
WHERE Packages.package_name = ? and Versions.version = ?
""",
                    (package_name, version),
                )

                for row in db_cursor.fetchall():
                    depends, timestamp, subdir, build = row
                    try:
                        lst = subdir_to_build_and_depends[subdir]
                    except KeyError:
                        lst = subdir_to_build_and_depends[subdir] = []

                    lst.append((build, depends))

                    max_timestamp = max(max_timestamp, timestamp)

            return CondaVersionInfo(
                package_name, version, max_timestamp, subdir_to_build_and_depends
            )


def version_key(version_info: CondaVersionInfo):
    from .conda_impl.conda_version import VersionOrder

    version_order = VersionOrder(version_info.version)
    return version_order


def sort_conda_version_infos(
    versions: Union[Sequence[CondaVersionInfo], Set[CondaVersionInfo]],
) -> List[CondaVersionInfo]:
    return list(sorted(versions, key=version_key))


def version_str_key(version: str):
    from .conda_impl.conda_version import VersionOrder

    version_order = VersionOrder(version)
    return version_order


def sort_conda_versions(
    versions: Union[Sequence[str], Set[str]],
) -> List[str]:
    return list(sorted(versions, key=version_str_key))


# Can be used in tests to index just a few libraries.
INDEX_FOR_LIBRARIES: Optional[Set[str]] = None


def index_conda_info(json_file: Path, target_sqlite_file: Path):
    import sqlite3

    # Using msgspec cuts the time in half (from 1 second to 0.5 seconds
    # for the full process in one use case).
    # Not only the decoding is faster (from 0.44 to 0.28), but accessing
    # things from the structure it provides is faster than accessing from
    # python structures (dict/list).
    import msgspec

    # Use Struct types to define the JSON schema. For efficiency we only define
    # the fields we actually need.
    class Package(msgspec.Struct):
        name: str
        depends: Optional[List[str]] = []
        version: Optional[str] = ""
        subdir: Optional[str] = ""
        timestamp: Optional[int] = 0
        build: Optional[str] = ""

    class Info(msgspec.Struct):
        subdir: str = ""

    class RepoData(msgspec.Struct):
        info: Info
        packages: Optional[Dict[str, Package]]
        packages_conda: Optional[Dict[str, Package]] = msgspec.field(
            name="packages.conda",
            default=None,
        )

    # Decode the data as a `RepoData` type
    with open(json_file, "rb") as stream:
        bytes_read = stream.read()
    data = msgspec.json.decode(bytes_read, type=RepoData)

    # This is actually the arch.
    default_subdir = data.info.subdir

    # Step 3: Create SQLite database and table
    db_connection = sqlite3.connect(target_sqlite_file)
    try:
        db_cursor = db_connection.cursor()
        try:
            db_cursor.execute(TABLE_PACKAGES_SQL)

            db_cursor.execute(TABLE_VERSIONS_SQL)

            for sql in CREATE_INDEXES_SQL:
                db_cursor.execute(sql)

            package_name_to_id: Dict[str, int] = {}
            rows = []
            # Step 4: Insert package information into the database
            for package_filename, package_info in itertools.chain(
                (data.packages or {}).items(), (data.packages_conda or {}).items()
            ):
                # "2dfatmic-1.0-hbb7d975_1.tar.bz2": {
                #   "build": "hbb7d975_1",
                #   "build": 1,
                #   "depends": [
                #     "libgfortran 5.*",
                #     "libgfortran5 >=9.3.0"
                #   ],
                #   "license": "Public Domain",
                #   "license_family": "OTHER",
                #   "md5": "0abb38856a2c0a45595bbf3888405507",
                #   "name": "2dfatmic",
                #   "sha256": "54d7427b37a4984e97c0529f94c927f9bccfbc9e5b57d0fb52030e4be92ce38a",
                #   "size": 173360,
                #   "subdir": "osx-64",
                #   "timestamp": 1602245771778,
                #   "version": "1.0"
                # },
                name = package_info.name
                if not name:
                    continue
                if INDEX_FOR_LIBRARIES:
                    if name not in INDEX_FOR_LIBRARIES:
                        continue

                try:
                    package_id = package_name_to_id[name]
                except KeyError:
                    db_cursor.execute(
                        """INSERT INTO Packages (package_name) VALUES (?);""", (name,)
                    )
                    lastrowid = db_cursor.lastrowid
                    if lastrowid is None:
                        continue
                    package_id = package_name_to_id[name] = lastrowid

                depends = msgspec.json.encode(package_info.depends)
                version = package_info.version
                timestamp = package_info.timestamp  # 0 means unknown
                build = package_info.build  # '' means unknown
                subdir = package_info.subdir or default_subdir

                if not version:
                    log.info(
                        f"Unable to get version for package_filename: {package_filename}. Subdir: {subdir}"
                    )
                    continue
                rows.append((package_id, depends, timestamp, version, subdir, build))

            db_cursor.executemany(
                "INSERT INTO Versions (package_id, depends, timestamp, version, subdir, build) VALUES (?, ?, ?, ?, ?, ?);",
                rows,
            )
        finally:
            db_cursor.close()

        db_connection.commit()
    finally:
        db_connection.close()


Arch = str


class _Callback(object):
    """
    Note that it's thread safe to register/unregister callbacks while callbacks
    are being notified, but it's not thread-safe to register/unregister at the
    same time in multiple threads.
    """

    def __init__(self):
        self._callbacks = []

    def register(self, callback):
        new_callbacks = self._callbacks[:]
        new_callbacks.append(callback)
        self._callbacks = new_callbacks

    def unregister(self, callback):
        new_callbacks = [x for x in self._callbacks if x != callback]
        self._callbacks = new_callbacks

    def __call__(self, *args, **kwargs):
        for c in self._callbacks:
            try:
                c(*args, **kwargs)
            except Exception:
                log.exception("Error in callback.")

    def __len__(self):
        return len(self._callbacks)

    def clear(self):
        self._callbacks = []


class _AfterDownloadMakeSqlite:
    def __init__(
        self, available_arch: List[Arch], cache_dir: Path, index_cache_dir: Path
    ):
        self._available_arch = available_arch
        self._lock = threading.Lock()
        self._count = 0
        self._call_on_finished = _Callback()
        self._finished = False
        self._arch_to_sqlite: Dict[str, Path] = {}
        self._index_cache_dir = index_cache_dir
        self._cache_dir = cache_dir

    def _convert_to_sqlite(self, json_file: Path, arch: Arch) -> Path:
        target_sqlite_file = self._index_cache_dir / f"{arch}.db"
        index_conda_info(json_file, target_sqlite_file)
        return target_sqlite_file

    def __call__(self, download_future: "Future[Tuple[Path, Arch]]"):
        json_file = None
        try:
            json_file, arch = download_future.result()
            self._arch_to_sqlite[arch] = self._convert_to_sqlite(json_file, arch)
        except BaseException:
            # Handle all exceptions (because the on_finished calls
            # must always be called).
            log.exception("Error when converting to sqlite.")

        try:
            if json_file is not None:
                # After we convert to sqlite, there's no need to keep the
                # (big) json file around anymore.
                os.remove(json_file)
        except Exception:
            log.exception(f"Error removing: {json_file}.")

        with self._lock:
            self._count += 1
            last = self._count == len(self._available_arch)
            if last:
                self._finished = True
                # If we're the last, change the information on the
                # cache dir to use.
                latest_path = self._cache_dir / "latest_index_info.json"
                latest_path.write_text(
                    json.dumps(
                        {
                            # Note: save relative to the cache dir.
                            "index_dir": os.path.basename(str(self._index_cache_dir)),
                            "timestamp": datetime.datetime.now().isoformat(),
                        }
                    ),
                    encoding="utf-8",
                )

                self._call_on_finished()

    def register_on_finished(self, on_finished: IOnFinished):
        with self._lock:
            if self._finished:
                on_finished()
            else:
                self._call_on_finished.register(on_finished)


def _extract_dir_number(dirname) -> int:
    try:
        _n, number = dirname.split("_")
        return int(number)
    except Exception:
        # Return zero if not what we expected.
        return 0


class CondaCloud:
    """
    Manages information from conda.

    The cache_dir is where we hold information on disk.

    The basic structure is:
        /cache_dir
        /cache_dir/tmp_0001 <-- used to download information and build the sqlite db.
        /cache_dir/tmp_0001/noarch.json <-- actually removed after the sqlite db is built.
        /cache_dir/tmp_0001/win-64.json
        /cache_dir/index_0001/noarch.db <-- sqlite db we actually use
        /cache_dir/index_0001/win-64.db
        /cache_dir/latest_index_info.json <-- will be used to point to index_0001
    """

    def __init__(self, cache_dir: Path, reindex_if_old: bool = True) -> None:
        self._base_url = "https://conda.anaconda.org/conda-forge"
        self._available_arch = [
            "noarch",
            "win-64",
            "linux-64",
            "osx-64",
        ]

        self._json_name = "current_repodata.json"

        self._cache_dir = cache_dir
        cache_dir.mkdir(parents=True, exist_ok=True)

        self._lock = threading.Lock()
        self._state: State = State.initial
        self._call_on_finished: List[IOnFinished] = []

        if self.is_information_cached():
            if reindex_if_old:
                # When a new CondaCloud is created, set it as done already if
                # 4 hours haven't elapsed from the last time it was requested.
                latest_index_info = self.load_latest_index_info()
                if latest_index_info is not None:
                    diff = datetime.datetime.now() - latest_index_info["timestamp"]
                    eight_hours = datetime.timedelta(hours=4)
                    if diff < eight_hours:
                        self._state = State.done
            else:
                self._state = State.done

    def _download(self, url: str, target_json: Path, arch: str) -> Tuple[Path, Arch]:
        import urllib.request

        CHUNK_SIZE = 32768

        with target_json.open("wb") as stream:
            request = urllib.request.urlopen(
                urllib.request.Request(url, headers={"User-Agent": "Mozilla"})
            )

            if request.status != 200:
                raise RuntimeError(
                    f"HTTP error (unable to open: {url}): {request.status}"
                )

            with request:
                for chunk in iter(lambda: request.read(CHUNK_SIZE), b""):
                    if chunk:  # Filter out keep-alive new chunks
                        stream.write(chunk)
        return target_json, arch

    def _call_on_finished_callbacks(self):
        for on_finished in self._call_on_finished:
            try:
                on_finished()
            except Exception:
                log.exception("Unexpected exception.")
        del self._call_on_finished[:]

    def schedule_update(
        self, on_finished: Optional[IOnFinished] = None, wait=False, force=False
    ) -> None:
        """
        This can be called at any time (in any thread). It schedules the download
        of the information we need.

        If it's currently downloading already it'll not do a new request and if it
        was already downloaded it'll only redownload if force == True.
        """
        import random

        with self._lock:
            if on_finished is not None:
                self._call_on_finished.append(on_finished)

            if self._state == State.initial:
                # Keep on going...
                pass

            elif self._state == State.downloading:
                # Ok, already downloading, nothing to do besides registering
                # the on_finished if given (already done)
                return

            elif self._state == State.done:
                if force:
                    # Forcing: go back to the initial state
                    self._state = State.initial
                else:
                    # Already finished, just call the registered callback now.
                    self._call_on_finished_callbacks()
                    return

            assert self._state in (State.initial, State.done)

            # We need to compute the target for the files
            dir_entries = os.listdir(self._cache_dir)

            stale_dirs: List[Path] = []
            max_tries = 3
            for i in range(max_tries):
                max_number: int = 0
                for dir_name in dir_entries:
                    if dir_name.startswith("tmp_") or dir_name.startswith("index_"):
                        max_number = max(max_number, _extract_dir_number(dir_name))

                for dir_name in dir_entries:
                    if dir_name.startswith("tmp_") or dir_name.startswith("index_"):
                        i = _extract_dir_number(dir_name)
                        if i < max_number:
                            stale_dirs.append(self._cache_dir / dir_name)

                max_number += 1

                tmp_cache_dir = self._cache_dir / ("tmp_%04d" % (max_number,))
                index_cache_dir = self._cache_dir / ("index_%04d" % (max_number,))
                try:
                    # If we failed to create the dir,
                    tmp_cache_dir.mkdir(parents=True, exist_ok=False)
                    index_cache_dir.mkdir(parents=True, exist_ok=False)
                except Exception:
                    if i == max_tries - 1:
                        raise
                    # Sleep a bit and retry, we could be running in parallel
                    # with another process (so, we may compute a new max_number).
                    time.sleep(random.random())
                else:
                    break

            # Just change the state when we have the directory structure in place.
            # (but still inside the self._lock session).
            self._state = State.downloading

            on_done = _AfterDownloadMakeSqlite(
                self._available_arch, self._cache_dir, index_cache_dir
            )

            def mark_as_done(*args, **kwargs):
                with self._lock:
                    self._state = State.done

                    # Now, remove stale dirs.
                    for directory in stale_dirs:
                        try:
                            shutil.rmtree(directory, ignore_errors=False)
                        except Exception:
                            log.exception(
                                f"Error removing old conda cloud dirs from: {directory}"
                            )

                    self._call_on_finished_callbacks()

            on_done.register_on_finished(mark_as_done)

            if on_finished:
                self._call_on_finished.append(on_finished)

        from concurrent.futures.thread import ThreadPoolExecutor

        executor = ThreadPoolExecutor(max_workers=4)

        for arch in self._available_arch:
            name = tmp_cache_dir / f"{arch}.json"
            # Something as:
            # https://conda.anaconda.org/conda-forge/linux-64/current_repodata.json
            url = f"{self._base_url}/{arch}/{self._json_name}"
            future = executor.submit(partial(self._download, url, name, arch))
            future.add_done_callback(on_done)
        executor.shutdown(wait=wait)

    def load_latest_index_info(self) -> Optional[LatestIndexInfoTypedDict]:
        try:
            latest_dir_info = self._cache_dir / "latest_index_info.json"
            with latest_dir_info.open("rb") as stream:
                info = json.load(stream)

            # index_dir is relative to the .json.
            index_dir_info: LatestIndexInfoTypedDict = {
                "index_dir": str(self._cache_dir / info["index_dir"]),
                "timestamp": datetime.datetime.fromisoformat(info["timestamp"]),
            }
            return index_dir_info
        except Exception:
            return None

    def _load_latest_index_dir_location(self) -> Optional[Path]:
        latest_dir_info = self.load_latest_index_info()
        if not latest_dir_info:
            return None

        try:
            index_dir_location = latest_dir_info["index_dir"]
        except Exception:
            return None

        index_dir_path = Path(index_dir_location)
        return index_dir_path

    def _iter_index_files(self) -> Iterator[Path]:
        """
        Returns:
            An iterator which provides the paths to the sqlite files
            which contain the indexed conda informations (one for each arch).
        """
        index_dir_path = self._load_latest_index_dir_location()
        if not index_dir_path or not index_dir_path.exists():
            return

        for arch in self._available_arch:
            f = index_dir_path / f"{arch}.db"
            if f.exists():
                yield f

    def is_information_cached(self) -> bool:
        index_dir_path = self._load_latest_index_dir_location()
        if not index_dir_path or not index_dir_path.exists():
            return False

        for arch in self._available_arch:
            if not (index_dir_path / f"{arch}.db").exists():
                return False

        return True

    def sqlite_queries(self) -> Optional[SqliteQueries]:
        index_dir_files = tuple(self._iter_index_files())
        if not index_dir_files:
            return None
        return SqliteQueries(index_dir_files)



# Original work Copyright (c) Django Software Foundation and individual contributors (BSD 3-Clause "New" or "Revised" License)
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Module based on: https://github.com/django/django/blob/main/django/utils/text.py
import re
import unicodedata


def slugify(value: str, allow_unicode: bool = False) -> str:
    """
    Convert to ASCII if 'allow_unicode' is False. Convert spaces or repeated
    dashes to single dashes. Remove characters that aren't alphanumerics,
    underscores, or hyphens. Convert to lowercase. Also strip leading and
    trailing whitespace, dashes, and underscores.
    """
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = _remove_accents(value).encode("ascii", "ignore").decode("ascii")

    value = re.sub(r"[^\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")


def _remove_accents(input_str):
    nfkd_form = unicodedata.normalize("NFKD", input_str)
    return "".join([c for c in nfkd_form if not unicodedata.combining(c)])



import datetime
import json
import logging
import time
import typing
from functools import partial
from typing import Any, Callable, Dict, Tuple

import fastjsonschema  # type: ignore
from fastapi import HTTPException, Request, Response
from fastapi.exceptions import RequestValidationError
from starlette.concurrency import run_in_threadpool

if typing.TYPE_CHECKING:
    from ._models import Action, Run

log = logging.getLogger(__name__)


# Note: for pydantic models, the following APIs are used:
# cls.model_validate(dict)
# cls.model_json_schema()
# obj.model_dump_json()
#
# Besides pydantic, the following basic types are accepted:
_spec_api_type_to_python_type = {
    "string": str,
    "integer": int,
    "number": float,
    "boolean": bool,
}


def _create_run_artifacts_dir(action: "Action", run_id: str) -> str:
    """
    Returns:
        The path, relative to the settings.artifacts_dir which should be used
        to store the output of the given run.
    """
    from ._settings import get_settings

    settings = get_settings()
    artifacts_dir = settings.artifacts_dir
    path = run_id
    target_dir = artifacts_dir / path
    target_dir.mkdir(parents=True, exist_ok=False)
    return path


def _create_run(
    action: "Action", run_id: str, inputs: dict, relative_artifacts_dir: str
) -> "Run":
    from robocorp.action_server._models import RUN_ID_COUNTER, Counter

    from ._database import datetime_to_str
    from ._models import Run, RunStatus, get_db
    from ._runs_state_cache import get_global_runs_state

    db = get_db()
    run_kwargs: Dict[str, Any] = dict(
        id=run_id,
        status=RunStatus.NOT_RUN,
        action_id=action.id,
        start_time=datetime_to_str(datetime.datetime.now(datetime.timezone.utc)),
        run_time=None,
        inputs=json.dumps(inputs),
        result=None,
        error_message=None,
        relative_artifacts_dir=relative_artifacts_dir,
    )
    with db.transaction():
        with db.cursor() as cursor:
            db.execute_update_returning(
                cursor,
                "UPDATE counter SET value=value+1 WHERE id=? RETURNING value",
                [RUN_ID_COUNTER],
            )
            counter_record = cursor.fetchall()
            if not counter_record:
                raise RuntimeError(
                    f"Error. No counter found for run_id. Counters in db: {db.all(Counter)}"
                )
            run_kwargs["numbered_id"] = counter_record[0][0]

        run = Run(**run_kwargs)
        db.insert(run)

    # Ok, transaction finished properly. Let's add it to our in-memory cache.
    global_runs_state = get_global_runs_state()
    global_runs_state.on_run_inserted(run)

    return run


def _update_run(run: "Run", initial_time: float, run_finished: bool, **changes):
    from ._models import get_db
    from ._runs_state_cache import get_global_runs_state

    if run_finished:
        changes["run_time"] = time.monotonic() - initial_time

    db = get_db()
    for k, v in changes.items():
        setattr(run, k, v)
    fields_changed = tuple(changes.keys())
    with db.transaction():
        db.update(run, *fields_changed)

    # Ok, transaction finished properly. Let's update our in-memory cache.
    global_runs_state = get_global_runs_state()
    global_runs_state.on_run_changed(run, changes)


def _set_run_as_finished_ok(run: "Run", result: str, initial_time: float):
    from ._models import RunStatus

    _update_run(run, initial_time, True, result=result, status=RunStatus.PASSED)


def _set_run_as_finished_failed(run: "Run", error_message: str, initial_time: float):
    from ._models import RunStatus

    _update_run(
        run, initial_time, True, status=RunStatus.FAILED, error_message=error_message
    )


def _set_run_as_running(run: "Run", initial_time: float):
    from ._models import RunStatus

    _update_run(run, initial_time, False, status=RunStatus.RUNNING)


def _run_action_in_thread(
    action: "Action",
    input_schema_dict: dict,
    output_schema_dict: dict,
    input_validator: Callable[[dict], None],
    output_validator: Callable[[dict], None],
    body: bytes,
    response: Response,
    request: Request,
):
    """
    This is where the user actually runs something.

    This runs in a thread (so, be careful when talking to the database).

    We have to take care of making a run with the proper environment,
    creating the run, collecting output info, etc.
    """
    from robocorp.action_server._gen_ids import gen_uuid
    from robocorp.action_server._settings import get_settings

    from ._actions_process_pool import get_actions_process_pool
    from ._models import Run, get_db

    settings = get_settings()

    try:
        inputs = json.loads(body)
    except Exception as e:
        raise RequestValidationError(
            [
                f"The received input arguments (sent in the body) cannot be interpreted as json. Details: {e}"
            ]
        )

    try:
        input_validator(inputs)
    except Exception as e:
        raise RequestValidationError(
            [
                f"The received input arguments (sent in the body) do not conform to the expected API. Details: {e}"
            ]
        )

    db = get_db()
    with db.connect():  # Connection is per-thread, so, we need to create a new one.
        actions_process_pool = get_actions_process_pool()
        reuse_process = settings.reuse_processes

        with actions_process_pool.obtain_process_for_action(action) as process_handle:
            run_id = gen_uuid("run")
            response.headers["X-Action-Server-Run-Id"] = run_id
            relative_artifacts_path: str = _create_run_artifacts_dir(action, run_id)
            run: Run = _create_run(action, run_id, inputs, relative_artifacts_path)

            input_json = (
                settings.artifacts_dir
                / relative_artifacts_path
                / "__action_server_inputs.json"
            )
            input_json.write_bytes(body)

            robot_artifacts = settings.artifacts_dir / relative_artifacts_path

            result_json = (
                settings.artifacts_dir
                / relative_artifacts_path
                / "__action_server_result.json"
            )

            output_file = (
                settings.artifacts_dir
                / relative_artifacts_path
                / "__action_server_output.txt"
            )

            initial_time = time.monotonic()
            try:
                _set_run_as_running(run, initial_time)
                returncode = process_handle.run_action(
                    action,
                    input_json,
                    robot_artifacts,
                    output_file,
                    result_json,
                    request,
                    reuse_process,
                )

                if returncode == 0:
                    try:
                        result: str = result_json.read_text("utf-8", "replace")
                    except Exception:
                        raise RuntimeError(
                            "It was not possible to collect the contents of the "
                            "result (json not created)."
                        )
                    try:
                        _set_run_as_finished_ok(run, result, initial_time)
                        ret = json.loads(result)
                    except Exception:
                        raise RuntimeError(
                            "Error loading the contents of {result_json} as json."
                        )
                    else:
                        try:
                            output_validator(ret)
                        except Exception as e:
                            raise RuntimeError(
                                f"Inconsistent value returned from action: {e} -- i.e.: the returned value ({ret!r}) does not match the expected output schema ({output_schema_dict})."
                            )
                        return ret

                raise RuntimeError(
                    f"Error: the process did not complete successfully when "
                    f"running. returncode: {returncode}"
                )

            except BaseException as e:
                _set_run_as_finished_failed(run, str(e), initial_time)
                raise HTTPException(status_code=500, detail=str(e))


def _name_as_class_name(name):
    return name.replace("_", " ").title().replace(" ", "")


def generate_func_from_action(
    action: "Action",
) -> Tuple[Callable[[Response, Request], Any], dict[str, object]]:
    """
    This method generates the function which should be called from FastAPI.

    Initially it generated the function with the parameters required to build
    the openapi.json spec properly, but it was changed to deal with the body
    directly and provide the openapi.json bits needed as it's easier and more
    straightforward to deal with the json directly and validate it than to
    create an in-memory python method with the proper signature so that FastAPI
    will do the validation.

    Returns:
        Function/Open API spec for the function
    """
    input_schema_dict = json.loads(action.input_schema)
    output_schema_dict = json.loads(action.output_schema)

    openapi_extra = {
        "requestBody": {
            "content": {
                "application/json": {
                    "schema": input_schema_dict,
                }
            },
            "required": True,
        },
        "responses": {
            "200": {
                "content": {
                    "application/json": {
                        "schema": {
                            **output_schema_dict,
                            **{
                                "title": f"Response {action.name.title().replace('_', ' ').replace('-', ' ')}"
                            },
                        }
                    }
                },
                "description": "Successful Response",
            },
            "422": {
                "content": {
                    "application/json": {
                        "schema": {"$ref": "#/components/schemas/HTTPValidationError"}
                    }
                },
                "description": "Validation Error",
            },
        },
    }

    try:
        input_validator = fastjsonschema.compile(input_schema_dict)
    except Exception:
        raise RuntimeError(
            f"Error making validator for input schema: {input_schema_dict}"
        )
    try:
        output_validator = fastjsonschema.compile(output_schema_dict)
    except Exception:
        raise RuntimeError(
            f"Errormaking validator for output schema: {output_schema_dict}"
        )

    # The returned function must be async because we have to request the `body`

    async def func(response: Response, request: Request):
        body = await request.body()

        return await run_in_threadpool(
            partial(
                _run_action_in_thread,
                action,
                input_schema_dict,
                output_schema_dict,
                input_validator,
                output_validator,
                body,
                response,
                request,
            )
        )

    return func, openapi_extra



"""
This package should be independent of the rest as we can potentially make
it a standalone package in the future (maybe with a command line UI).
"""

import pathlib
import typing
from typing import Any, Iterator, List, Optional, Tuple, Union

import yaml

from ._deps_protocols import (
    ICondaCloud,
    IPyPiCloud,
    _DiagnosticSeverity,
    _DiagnosticsTypedDict,
    _RangeTypedDict,
)

if typing.TYPE_CHECKING:
    from ._conda_deps import CondaDepInfo, CondaDeps
    from ._package_deps import PackageDepsConda, PackageDepsPip
    from ._pip_deps import PipDepInfo, PipDeps


class ScalarInfo:
    def __init__(self, scalar: Any, location: Optional[Tuple[int, int, int, int]]):
        """
        Args:
            scalar:
            location: tuple(start_line, start_col, end_line, end_col)
        """
        self.scalar = scalar
        self.location = location

    def __repr__(self):
        return repr(self.scalar)

    def __str__(self):
        return str(self.scalar)

    def __eq__(self, o):
        if isinstance(o, ScalarInfo):
            return self.scalar == o.scalar

        return False

    def __ne__(self, o):
        return not self == o

    def __hash__(self):
        return hash(self.scalar)

    def as_range(self) -> _RangeTypedDict:
        assert self.location
        start_line, start_col, end_line, end_col = self.location
        return create_range_from_location(start_line, start_col, end_line, end_col)


def create_range_from_location(
    start_line: int,
    start_col: int,
    end_line: Optional[int] = None,
    end_col: Optional[int] = None,
) -> _RangeTypedDict:
    """
    If the end_line and end_col aren't passed we consider
    that the location should go up until the end of the line.
    """
    if end_line is None:
        assert end_col is None
        end_line = start_line + 1
        end_col = 0
    assert end_col is not None
    dct: _RangeTypedDict = {
        "start": {
            "line": start_line,
            "character": start_col,
        },
        "end": {
            "line": end_line,
            "character": end_col,
        },
    }
    return dct


class LoaderWithLines(yaml.SafeLoader):
    def construct_scalar(self, node, *args, **kwargs):
        scalar = yaml.SafeLoader.construct_scalar(self, node, *args, **kwargs)
        return ScalarInfo(
            scalar=scalar,
            location=(
                node.start_mark.line,
                node.start_mark.column,
                node.end_mark.line,
                node.end_mark.column,
            ),
        )


class BaseAnalyzer:
    _pypi_cloud: IPyPiCloud
    _conda_cloud: ICondaCloud

    _pip_deps: "Union[PipDeps, PackageDepsPip]"
    _conda_deps: "Union[CondaDeps, PackageDepsConda]"

    def __init__(
        self,
        contents: str,
        path: str,
        conda_cloud: ICondaCloud,
        pypi_cloud: Optional[IPyPiCloud] = None,
    ):
        """
        Args:
            contents: The contents of the conda.yaml/action-server.yaml.
            path: The path for the conda yaml.
        """
        from .pypi_cloud import PyPiCloud

        self.contents = contents
        self.path = path

        self._loaded_yaml = False
        self._load_errors: List[_DiagnosticsTypedDict] = []
        self._yaml_data: Optional[dict] = None

        if pypi_cloud is None:
            self._pypi_cloud = PyPiCloud()
        else:
            self._pypi_cloud = pypi_cloud

        self._conda_cloud = conda_cloud

    def _load_yaml_info(self) -> None:
        """
        Loads the base information and creates errors for syntax errors
        loading the yaml.
        """
        if self._loaded_yaml:
            return
        self._loaded_yaml = True

        from yaml.parser import ParserError

        diagnostic: _DiagnosticsTypedDict

        try:
            loader = LoaderWithLines(self.contents)
            path: pathlib.Path = pathlib.Path(self.path)

            loader.name = f".../{path.parent.name}/{path.name}"
            data = loader.get_single_data()
            if isinstance(data, dict):
                self._yaml_data = data
            else:
                diagnostic = {
                    "range": create_range_from_location(0, 0),
                    "severity": _DiagnosticSeverity.Error,
                    "source": "robocorp-code",
                    "message": "Error: expected dict to be root of yaml",
                }
                self._load_errors.append(diagnostic)

        except ParserError as e:
            if e.problem_mark:
                diagnostic = {
                    "range": create_range_from_location(
                        e.problem_mark.line, e.problem_mark.column
                    ),
                    "severity": _DiagnosticSeverity.Error,
                    "source": "robocorp-code",
                    "message": f"Syntax error: {e}",
                }
                self._load_errors.append(diagnostic)

    def iter_pip_issues(self):
        from .pip_impl import pip_packaging_version

        for dep_info in self._pip_deps.iter_deps_infos():
            if dep_info.error_msg:
                diagnostic = {
                    "range": dep_info.dep_range,
                    "severity": _DiagnosticSeverity.Error,
                    "source": "robocorp-code",
                    "message": dep_info.error_msg,
                }

                yield diagnostic

                continue

            if dep_info.constraints and len(dep_info.constraints) == 1:
                # For now just checking versions '=='.
                constraint = next(iter(dep_info.constraints))
                if constraint[0] == "==":
                    local_version = constraint[1]
                    newer_cloud_versions = self._pypi_cloud.get_versions_newer_than(
                        dep_info.name, pip_packaging_version.parse(local_version)
                    )
                    if newer_cloud_versions:
                        latest_cloud_version = newer_cloud_versions[-1]
                        diagnostic = {
                            "range": dep_info.dep_range,
                            "severity": _DiagnosticSeverity.Warning,
                            "source": "robocorp-code",
                            "message": f"Consider updating '{dep_info.name}' to the latest version: '{latest_cloud_version}'. "
                            + f"Found {len(newer_cloud_versions)} versions newer than the current one: {', '.join(newer_cloud_versions)}.",
                        }

                        yield diagnostic

    def iter_conda_issues(self) -> Iterator[_DiagnosticsTypedDict]:
        from .conda_cloud import sort_conda_versions
        from .conda_impl.conda_version import VersionSpec

        diagnostic: _DiagnosticsTypedDict
        dep_vspec = self._conda_deps.get_dep_vspec("python")

        # See: https://devguide.python.org/versions/

        if dep_vspec is not None:
            # We have the dep spec, not the actual version
            # (so, it could be something as >3.3 <4)
            # So, if an old version may be gotten from that spec, we warn about it.
            vspec = VersionSpec(dep_vspec)
            for old_version in "2 3.1 3.2 3.3 3.4 3.5 3.6 3.7".split(" "):
                if vspec.match(old_version):
                    dep_range = self._conda_deps.get_dep_range("python")

                    diagnostic = {
                        "range": dep_range,
                        "severity": _DiagnosticSeverity.Warning,
                        "source": "robocorp-code",
                        "message": "The official support for versions lower than Python 3.8 has ended. "
                        + " It is advisable to transition to a newer version "
                        + "(Python 3.9 or newer is recommended).",
                    }

                    yield diagnostic
                    break

        dep_vspec = self._conda_deps.get_dep_vspec("pip")

        if dep_vspec is not None:
            vspec = VersionSpec(dep_vspec)
            # pip should be 22 or newer, so, check if an older version matches.
            for old_version in (str(x) for x in range(1, 22)):
                if vspec.match(old_version):
                    dep_range = self._conda_deps.get_dep_range("pip")

                    diagnostic = {
                        "range": dep_range,
                        "severity": _DiagnosticSeverity.Warning,
                        "source": "robocorp-code",
                        "message": "Consider updating pip to a newer version (at least pip 22 onwards is recommended).",
                    }

                    yield diagnostic
                    break

        sqlite_queries = self._conda_cloud.sqlite_queries()
        if sqlite_queries:
            with sqlite_queries.db_cursors() as db_cursors:
                for conda_dep in self._conda_deps.iter_deps_infos():
                    if conda_dep.name in ("python", "pip"):
                        continue

                    if conda_dep.error_msg:
                        diagnostic = {
                            "range": conda_dep.dep_range,
                            "severity": _DiagnosticSeverity.Error,
                            "source": "robocorp-code",
                            "message": conda_dep.error_msg,
                        }

                        yield diagnostic

                        continue

                    version_spec = conda_dep.get_dep_vspec()
                    if version_spec is None:
                        continue

                    versions = sqlite_queries.query_versions(conda_dep.name, db_cursors)
                    if not versions:
                        continue

                    sorted_versions = sort_conda_versions(versions)
                    last_version = sorted_versions[-1]
                    if not version_spec.match(last_version):
                        # The latest version doesn't match, let's show a warning.
                        newer_cloud_versions = []
                        for v in reversed(sorted_versions):
                            if not version_spec.match(v):
                                newer_cloud_versions.append(v)
                            else:
                                break

                        diagnostic = {
                            "range": conda_dep.dep_range,
                            "severity": _DiagnosticSeverity.Warning,
                            "source": "robocorp-code",
                            "message": f"Consider updating '{conda_dep.name}' to match the latest version: '{last_version}'. "
                            + f"Found {len(newer_cloud_versions)} versions that don't match the version specification: {', '.join(newer_cloud_versions)}.",
                        }

                        yield diagnostic

    def find_pip_dep_at(self, line, col) -> "Optional[PipDepInfo]":
        """
        Args:
            line: 0-based line
        """
        self._load_yaml_info()
        for dep_info in self._pip_deps.iter_deps_infos():
            if is_inside(dep_info.dep_range, line, col):
                return dep_info
        return None

    def find_conda_dep_at(self, line, col) -> "Optional[CondaDepInfo]":
        self._load_yaml_info()
        for dep_info in self._conda_deps.iter_deps_infos():
            if is_inside(dep_info.dep_range, line, col):
                return dep_info
        return None


class PackageYamlAnalyzer(BaseAnalyzer):
    def __init__(
        self,
        contents: str,
        path: str,
        conda_cloud: ICondaCloud,
        pypi_cloud: Optional[IPyPiCloud] = None,
    ):
        from ._package_deps import PackageDepsConda, PackageDepsPip

        self._pip_deps = PackageDepsPip()
        self._conda_deps = PackageDepsConda()
        self._additional_load_errors: List[_DiagnosticsTypedDict] = []

        BaseAnalyzer.__init__(self, contents, path, conda_cloud, pypi_cloud=pypi_cloud)

    def iter_package_yaml_issues(self) -> Iterator[_DiagnosticsTypedDict]:
        self._load_yaml_info()
        if self._load_errors:
            yield from iter(self._load_errors)
            return

        data = self._yaml_data
        if not data:
            return

        yield from iter(self._additional_load_errors)
        yield from self.iter_conda_issues()
        yield from self.iter_pip_issues()

    def _load_yaml_info(self) -> None:
        if self._loaded_yaml:
            return
        BaseAnalyzer._load_yaml_info(self)
        data = self._yaml_data
        if not data:
            return

        diagnostic: _DiagnosticsTypedDict
        dependencies_key_entry: ScalarInfo
        for key, v in data.items():
            if isinstance(key, ScalarInfo) and isinstance(key.scalar, str):
                if key.scalar == "dependencies":
                    dependencies = v
                    dependencies_key_entry = key
                    break
        else:
            diagnostic = {
                "range": create_range_from_location(0, 0),
                "severity": _DiagnosticSeverity.Error,
                "source": "robocorp-code",
                "message": "Error: 'dependencies' entry not found",
            }
            self._additional_load_errors.append(diagnostic)
            return

        if not dependencies:
            diagnostic = {
                "range": create_range_from_location(0, 0),
                "severity": _DiagnosticSeverity.Error,
                "source": "robocorp-code",
                "message": "Error: 'dependencies' entry must not be empty",
            }
            self._additional_load_errors.append(diagnostic)

        elif not isinstance(dependencies, dict):
            diagnostic = {
                "range": create_range_from_location(0, 0),
                "severity": _DiagnosticSeverity.Error,
                "source": "robocorp-code",
                "message": f"Error: expected 'dependencies' entry to be a dict (with 'conda-forge' and 'pypi' entries). Found: '{type(dependencies)}'",
            }
            self._additional_load_errors.append(diagnostic)
        else:
            for dep_name, dep in dependencies.items():
                if isinstance(dep_name, ScalarInfo):
                    if dep_name.scalar == "pypi":
                        if not isinstance(dep, list):
                            diagnostic = {
                                "range": dep_name.as_range(),
                                "severity": _DiagnosticSeverity.Error,
                                "source": "robocorp-code",
                                "message": f"Error: expected the entries of pypi to be a list. Found: {type(dep)}",
                            }
                            self._additional_load_errors.append(diagnostic)
                            continue

                        for entry in dep:
                            if not isinstance(entry, ScalarInfo):
                                diagnostic = {
                                    "range": dep_name.as_range(),
                                    "severity": _DiagnosticSeverity.Error,
                                    "source": "robocorp-code",
                                    "message": f"Error: expected the entries of pypi to be strings. Found: {type(entry)}: {entry}",
                                }
                                self._additional_load_errors.append(diagnostic)
                                continue

                            if (
                                entry.scalar.replace(" ", "")
                                == "--use-feature=truststore"
                            ):
                                diagnostic = {
                                    "range": entry.as_range(),
                                    "severity": _DiagnosticSeverity.Warning,
                                    "source": "robocorp-code",
                                    "message": (
                                        "--use-feature=truststore flag does not need to "
                                        "be specified (it is automatically used when a "
                                        '"robocorp-trustore" or "trustore" dependency is added).'
                                    ),
                                }
                                self._additional_load_errors.append(diagnostic)
                                continue

                            if entry.scalar.startswith("--"):
                                diagnostic = {
                                    "range": entry.as_range(),
                                    "severity": _DiagnosticSeverity.Error,
                                    "source": "robocorp-code",
                                    "message": f"Invalid entry in pypi: {entry.scalar}",
                                }
                                self._additional_load_errors.append(diagnostic)
                                continue

                            self._pip_deps.add_dep(entry.scalar, entry.as_range())

                    elif dep_name.scalar == "conda-forge":
                        for entry in dep:
                            if not isinstance(entry, ScalarInfo):
                                diagnostic = {
                                    "range": dep_name.as_range(),
                                    "severity": _DiagnosticSeverity.Error,
                                    "source": "robocorp-code",
                                    "message": f"Error: expected the entries of conda-forge to be strings. Found: {type(entry)}: {entry}",
                                }
                                self._additional_load_errors.append(diagnostic)
                                continue

                            self._conda_deps.add_dep(entry.scalar, entry.as_range())

                    else:
                        diagnostic = {
                            "range": dep_name.as_range(),
                            "severity": _DiagnosticSeverity.Error,
                            "source": "robocorp-code",
                            "message": (
                                "Error: only expected children are pypi and conda (dict entries). "
                                f"Found: {dep_name.scalar}"
                            ),
                        }
                        self._additional_load_errors.append(diagnostic)
                else:
                    diagnostic = {
                        "range": dependencies_key_entry.as_range(),
                        "severity": _DiagnosticSeverity.Error,
                        "source": "robocorp-code",
                        "message": "Error: found unexpected entry in dependencies.",
                    }
                    self._additional_load_errors.append(diagnostic)


class CondaYamlAnalyzer(BaseAnalyzer):
    def __init__(
        self,
        contents: str,
        path: str,
        conda_cloud: ICondaCloud,
        pypi_cloud: Optional[IPyPiCloud] = None,
    ):
        from ._conda_deps import CondaDeps
        from ._pip_deps import PipDeps

        self._pip_deps = PipDeps()
        self._conda_deps = CondaDeps()

        BaseAnalyzer.__init__(self, contents, path, conda_cloud, pypi_cloud=pypi_cloud)

    def _load_yaml_info(self) -> None:
        if self._loaded_yaml:
            return
        BaseAnalyzer._load_yaml_info(self)
        data = self._yaml_data
        if not data:
            return

        dependencies = data.get(
            ScalarInfo(
                "dependencies",
                None,
            ),
        )

        conda_versions = self._conda_deps
        pip_versions = self._pip_deps
        if dependencies:
            for dep in dependencies:
                # At this level we're seeing conda versions. The spec is:
                # https://docs.conda.io/projects/conda-build/en/latest/resources/package-spec.html#package-match-specifications
                # A bunch of code from conda was copied to handle that so that we
                # can just `conda_match_spec.parse_spec_str` to identify the version
                # we're dealing with.
                if isinstance(dep, ScalarInfo) and isinstance(dep.scalar, str):
                    conda_versions.add_dep(dep.scalar, dep.as_range())
                elif isinstance(dep, dict):
                    pip_deps = dep.get(ScalarInfo("pip", None))
                    if pip_deps:
                        for dep in pip_deps:
                            if isinstance(dep, ScalarInfo) and isinstance(
                                dep.scalar, str
                            ):
                                pip_versions.add_dep(dep.scalar, dep.as_range())

    def iter_conda_yaml_issues(self) -> Iterator[_DiagnosticsTypedDict]:
        self._load_yaml_info()
        if self._load_errors:
            yield from iter(self._load_errors)
            return

        data = self._yaml_data
        if not data:
            return

        yield from self.iter_conda_issues()
        yield from self.iter_pip_issues()


class _Position:
    def __init__(self, line: int = 0, character: int = 0):
        self.line: int = line
        self.character: int = character

    def to_dict(self):
        return {"line": self.line, "character": self.character}

    def __repr__(self):
        import json

        return json.dumps(self.to_dict(), indent=4)

    def __getitem__(self, name):
        # provide tuple-access, not just dict access.
        if name == 0:
            return self.line
        if name == 1:
            return self.character
        return getattr(self, name)

    def __eq__(self, other):
        return (
            isinstance(other, _Position)
            and self.line == other.line
            and self.character == other.character
        )

    def __ge__(self, other):
        line_gt = self.line > other.line

        if line_gt:
            return line_gt

        if self.line == other.line:
            return self.character >= other.character

        return False

    def __gt__(self, other):
        line_gt = self.line > other.line

        if line_gt:
            return line_gt

        if self.line == other.line:
            return self.character > other.character

        return False

    def __le__(self, other):
        line_lt = self.line < other.line

        if line_lt:
            return line_lt

        if self.line == other.line:
            return self.character <= other.character

        return False

    def __lt__(self, other):
        line_lt = self.line < other.line

        if line_lt:
            return line_lt

        if self.line == other.line:
            return self.character < other.character

        return False

    def __ne__(self, other):
        return not self.__eq__(other)


def is_inside(range_dct: _RangeTypedDict, line: int, col: int) -> bool:
    start = range_dct["start"]
    end = range_dct["end"]
    start_pos = _Position(start["line"], start["character"])
    end_pos = _Position(end["line"], end["character"])
    curr_pos = _Position(line, col)
    return start_pos <= curr_pos <= end_pos



try:
    from termcolor import colored
except ImportError:

    def _noop(msg):
        return msg

    bold_yellow = _noop
    bold_red = _noop
else:

    def bold_yellow(msg):
        return colored(msg, color="yellow", attrs=["bold"])

    def bold_red(msg):
        return colored(msg, color="red", attrs=["bold"])



from dataclasses import dataclass
from typing import Any, Iterator, List, Optional, Tuple

from ._deps_protocols import _RangeTypedDict


@dataclass
class PipDepInfo:
    name: str  # rpaframework
    extras: Any
    constraints: Optional[List[Tuple[str, str]]]  # i.e.: [('==', '22.5.3')]
    marker: Any
    url: str
    requirement: str  #'rpaframework == 22.5.3'
    dep_range: _RangeTypedDict
    error_msg: Optional[str]  # Error if we couldn't parse it.


class PipDeps:
    """
    pip references:
        pip._vendor.distlib.version.Matcher
        pip._vendor.distlib.util.parse_requirement
    """

    def __init__(self):
        self._deps = {}

    def add_dep(self, value, as_range):
        from .pip_impl.pip_distlib_util import IDENTIFIER, parse_requirement

        try:
            req = parse_requirement(value)
        except Exception as e:
            remaining = value.strip()
            # -- ignored because it's used for pip flags (i.e.: --use-feature=trustore).
            if not remaining or remaining.startswith("#") or remaining.startswith("--"):
                return

            if value.startswith(("git+http", "http")) or value.endswith(
                (".whl", ".zip", "bz2", "gz")
            ):
                self._deps[value] = PipDepInfo(
                    name=value,
                    extras=None,
                    constraints=None,
                    marker=None,
                    url="",
                    requirement=value,
                    dep_range=as_range,
                    error_msg=None,
                )
                return

            m = IDENTIFIER.match(remaining)
            if not m:
                name = "<unknown>"
            else:
                name = m.groups()[0]

            self._deps[name] = PipDepInfo(
                name=name,
                extras=None,
                constraints=None,
                marker=None,
                url="",
                requirement=value,
                dep_range=as_range,
                error_msg=f"Error parsing '{value}': {e}",
            )
        else:
            self._deps[req.name] = PipDepInfo(
                **req.__dict__, error_msg=None, dep_range=as_range
            )

    def iter_deps_infos(self) -> Iterator[PipDepInfo]:
        yield from self._deps.values()



import datetime
import logging
import typing
import weakref
from typing import Dict, Iterator, List, Optional, Sequence, Union

from ._deps_protocols import PyPiInfoTypedDict, ReleaseData, Versions, VersionStr

log = logging.getLogger(__name__)

# Interesting:
# https://github.com/python-poetry/poetry/blob/master/src/poetry/repositories/pypi_repository.py


class PackageData:
    def __init__(self, package_name: str, info: PyPiInfoTypedDict) -> None:
        self.package_name = package_name
        self._info: PyPiInfoTypedDict = info

        self._releases: Dict[str, ReleaseData] = {}

    def add_release(self, version_str: VersionStr, release_info: List[dict]) -> None:
        """
        Args:
            version_str: The version we have info on.
            release_info: For each release we may have a list of files available.
        """
        from .pip_impl import pip_packaging_version

        version = pip_packaging_version.parse(version_str)
        upload_time: Optional[str] = None
        for dct in release_info:
            upload_time = dct.get("upload_time")
            if upload_time:
                break

        self._releases[version_str] = ReleaseData(version, version_str, upload_time)

    @property
    def latest_version(self) -> VersionStr:
        return self._info["version"]

    def get_release_data(self, version: VersionStr) -> Optional[ReleaseData]:
        if not self._releases:
            return None
        try:
            return self._releases[version]
        except KeyError:
            return None

    def get_last_release_data(self) -> Optional[ReleaseData]:
        """
        Provides the last release data (if there's any release).
        """
        return self.get_release_data(self.latest_version)

    def iter_versions_released_after(
        self, after_datetime: Optional[datetime.datetime]
    ) -> Iterator[ReleaseData]:
        """
        Args:
            after_datetime: if none all releases (except pre-releases) will
            be provided.
        """

        last_release_data = self.get_last_release_data()
        latest_upload_datetime: Optional[datetime.datetime] = None
        if last_release_data and last_release_data.upload_time:
            latest_upload_datetime = datetime.datetime.strptime(
                last_release_data.upload_time, "%Y-%m-%dT%H:%M:%S"
            )

        for release_data in self._releases.values():
            if release_data.upload_time:
                upload_datetime = datetime.datetime.strptime(
                    release_data.upload_time, "%Y-%m-%dT%H:%M:%S"
                )

                if latest_upload_datetime:
                    # Hide pre-releases.
                    if upload_datetime > latest_upload_datetime:
                        continue

                if after_datetime is None or upload_datetime >= after_datetime:
                    yield release_data

    def iter_versions_newer_than(self, version: Versions) -> Iterator[ReleaseData]:
        for release_data in self.iter_versions_released_after(None):
            if release_data.version > version:
                yield release_data

    @property
    def info(self):
        return self._info


class PyPiCloud:
    def __init__(
        self, get_base_urls_weak_method: Optional[weakref.WeakMethod] = None
    ) -> None:
        self._cached_package_data: Dict[str, PackageData] = {}
        self._cached_cloud: Dict[str, PackageData] = {}
        self._last_base_urls: Sequence[str] = ()

        if get_base_urls_weak_method is None:
            # use pypi.org
            def get_weak():
                def get_pypi_url():
                    return ("https://pypi.org",)

                return get_pypi_url

            self.get_base_urls_weak_method = get_weak
        else:
            self.get_base_urls_weak_method = get_base_urls_weak_method

    def _get_json_from_cloud(self, url: str) -> Optional[dict]:
        try:
            return typing.cast(dict, self._cached_cloud[url])
        except KeyError:
            pass

        import json
        import urllib.request

        try:
            request = urllib.request.urlopen(
                urllib.request.Request(url, headers={"User-Agent": "Mozilla"})
            )
            if request.status == 200:
                with request:
                    data = request.read().decode("utf-8", "replace")
                self._cached_cloud[url] = json.loads(data)
            else:
                log.info(
                    f"Unable to get url (as json): {url}. Status code: {request.status}"
                )
                return None
        except Exception as e:
            log.info(f"Unable to get url (as json): {url}. Error: {e}")
            return None

        return typing.cast(dict, self._cached_cloud[url])

    def get_package_data(self, package_name: str) -> Optional[PackageData]:
        get_base_urls = self.get_base_urls_weak_method()
        if get_base_urls is None:
            return None

        base_urls = get_base_urls()
        if base_urls != self._last_base_urls:
            # We need to clear packages if urls changed.
            self._last_base_urls = base_urls
            self._cached_package_data.clear()
            self._cached_cloud.clear()

        try:
            return self._cached_package_data[package_name]
        except KeyError:
            pass

        for base_url in base_urls:
            if base_url.endswith("/"):
                base_url = base_url[:-1]
            data = self._get_json_from_cloud(f"{base_url}/pypi/{package_name}/json")
            if not data:
                continue  # go to the next url

            try:
                releases = data["releases"]
            except KeyError:
                continue  # go to the next url

            try:
                info = data["info"]
            except KeyError:
                continue  # go to the next url

            package_data = PackageData(package_name, info)
            if releases and isinstance(releases, dict):
                for release_number, release_info in releases.items():
                    package_data.add_release(release_number, release_info)

            self._cached_package_data[package_name] = package_data
            return self._cached_package_data[package_name]

        # If there was no match, return.
        return None

    def get_versions_newer_than(
        self, package_name: str, version: Union[Versions, VersionStr]
    ) -> List[VersionStr]:
        """
        Args:
            package_name: The name of the package
            version: The minimum version (versions returned must be > than this one).

        Returns:
            A sorted list containing the versions > than the one passed (the last
            entry is the latest version).
        """

        if isinstance(version, VersionStr):
            from .pip_impl import pip_packaging_version

            version = pip_packaging_version.parse(version)

        package_data = self.get_package_data(package_name)
        if package_data is None:
            return []

        return [
            x.version_str
            for x in sorted(package_data.iter_versions_newer_than(version))
        ]



class Channel:
    """
    Channel:
    scheme <> auth <> location <> token <> channel <> subchannel <> platform <> package_filename

    Package Spec:
    channel <> subchannel <> namespace <> package_name

    """

    def __init__(
        self,
        scheme=None,
        auth=None,
        location=None,
        token=None,
        name=None,
        platform=None,
        package_filename=None,
    ):
        self.scheme = scheme
        self.auth = auth
        self.location = location
        self.token = token
        self.name = name or ""
        self.platform = platform
        self.package_filename = package_filename

    @property
    def channel_location(self):
        return self.location

    @property
    def channel_name(self):
        return self.name

    @property
    def subdir(self):
        return self.platform



# Versions = Union[LegacyVersion, Version]
import datetime
import sys
import typing
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Dict, Iterator, List, Optional, Sequence, Set, Tuple, Union

# Hack so that we don't break the runtime on versions prior to Python 3.8.
if sys.version_info[:2] < (3, 8):

    class Protocol(object):
        pass

    class TypedDict(object):
        pass

else:
    from typing import Protocol, TypedDict


VersionStr = str


class Versions(Protocol):
    def __lt__(self, other):
        pass


@dataclass
class ReleaseData:
    version: Versions
    version_str: VersionStr
    upload_time: Optional[str]  # "2020-02-05T14:11:38"

    def __lt__(self, other):
        return self.version < other.version


UrlDescription = str
UrlStr = str


class PyPiInfoTypedDict(TypedDict):
    description: str
    description_content_type: str
    home_page: str
    package_url: str
    project_urls: Dict[UrlDescription, UrlStr]

    # Info for the latest version
    # The constraints. i.e.: ['rpaframework-windows (>=7.3.2,<8.0.0) ; sys_platform == "win32"']
    # Something as ">=3.7,<4.0"
    requires_dist: List[str]
    requires_python: str
    version: VersionStr


class IPackageData(Protocol):
    package_name: str

    def add_release(self, version_str: VersionStr, release_info: List[dict]) -> None:
        """
        Args:
            version_str: The version we have info on.
            release_info: For each release we may have a list of files available.
        """

    @property
    def latest_version(self) -> VersionStr:
        pass

    def get_last_release_data(self) -> Optional[ReleaseData]:
        pass

    def iter_versions_released_after(
        self, after_datetime: datetime.datetime
    ) -> Iterator[ReleaseData]:
        pass

    def iter_versions_newer_than(self, version: Versions) -> Iterator[ReleaseData]:
        pass

    @property
    def info(self):
        pass

    def get_release_data(self, version: VersionStr) -> Optional[ReleaseData]:
        pass


class IPyPiCloud(Protocol):
    def get_package_data(self, package_name: str) -> Optional[IPackageData]:
        pass

    def get_versions_newer_than(
        self, package_name: str, version: Union[Versions, VersionStr]
    ) -> List[VersionStr]:
        """
        Args:
            package_name: The name of the package
            version: The minimum version (versions returned must be > than this one).

        Returns:
            A sorted list containing the versions > than the one passed (the last
            entry is the latest version).
        """


SubdirToBuildAndDependsJsonBytesType = Dict[str, List[Tuple[str, bytes]]]


@dataclass(unsafe_hash=True)
class CondaVersionInfo:
    package_name: str
    version: str  # This is the version from conda-forge.
    timestamp: int  # Max is given
    subdir_to_build_and_depends_json_bytes: SubdirToBuildAndDependsJsonBytesType


if typing.TYPE_CHECKING:
    # No need for sqlite3 until it's actually used.
    from sqlite3 import Cursor
else:
    Cursor = object


class LatestIndexInfoTypedDict(TypedDict):
    # The place where the index info is saved.
    index_dir: str
    timestamp: datetime.datetime


class ISqliteQueries(Protocol):
    @contextmanager
    def db_cursors(
        self, db_cursor: Optional[Sequence[Cursor]] = None
    ) -> Iterator[Sequence[Cursor]]:
        pass

    def query_names(self, db_cursors: Optional[Sequence[Cursor]] = None) -> Set[str]:
        pass

    def query_versions(
        self, package_name, db_cursors: Optional[Sequence[Cursor]] = None
    ) -> Set[str]:
        pass

    def query_version_info(
        self,
        package_name: str,
        version: str,
        db_cursors: Optional[Sequence[Cursor]] = None,
    ) -> CondaVersionInfo:
        pass


class IOnFinished(Protocol):
    def __call__(self):
        pass


class ICondaCloud(Protocol):
    def is_information_cached(self) -> bool:
        pass

    def sqlite_queries(self) -> Optional[ISqliteQueries]:
        pass

    def schedule_update(
        self, on_finished: Optional[IOnFinished] = None, wait=False, force=False
    ) -> None:
        pass


class _DiagnosticSeverity(object):
    Error = 1
    Warning = 2
    Information = 3
    Hint = 4


class _PositionTypedDict(TypedDict):
    # Line position in a document (zero-based).
    line: int

    # Character offset on a line in a document (zero-based). Assuming that
    # the line is represented as a string, the `character` value represents
    # the gap between the `character` and `character + 1`.
    #
    # If the character value is greater than the line length it defaults back
    # to the line length.
    character: int


class _RangeTypedDict(TypedDict):
    start: _PositionTypedDict
    end: _PositionTypedDict


class _DiagnosticsTypedDict(TypedDict, total=False):
    # The range at which the message applies.
    range: _RangeTypedDict

    # The diagnostic's severity. Can be omitted. If omitted it is up to the
    # client to interpret diagnostics as error, warning, info or hint.
    severity: Optional[int]  # DiagnosticSeverity

    # The diagnostic's code, which might appear in the user interface.
    code: Union[int, str]

    # An optional property to describe the error code.
    #
    # @since 3.16.0
    codeDescription: Any

    # A human-readable string describing the source of this
    # diagnostic, e.g. 'typescript' or 'super lint'.
    source: Optional[str]

    # The diagnostic's message.
    message: str

    # Additional metadata about the diagnostic.
    #
    # @since 3.15.0
    tags: list  # DiagnosticTag[];

    # An array of related diagnostic information, e.g. when symbol-names within
    # a scope collide all definitions can be marked via this property.
    relatedInformation: list  # DiagnosticRelatedInformation[];

    # A data entry field that is preserved between a
    # `textDocument/publishDiagnostics` notification and
    # `textDocument/codeAction` request.
    #
    # @since 3.16.0
    data: Optional[Any]  # unknown;



"""
This module can be used to run the actions in a way that the actions are pre-loaded
(i.e.: imported) but no actions are actually run until requested by the action-server.

Important: this will run in the target environment and can't really import anything
from the action server.
"""
import argparse
import os
import sys
import traceback
from typing import Any, Dict

DEFAULT_TIMEOUT = 10
NO_TIMEOUT = None
USE_TIMEOUTS = True


class _DummyStdin(object):
    def __init__(self, original_stdin=sys.stdin, *args, **kwargs):
        try:
            self.encoding = sys.stdin.encoding
        except Exception:
            # Not sure if it's available in all Python versions...
            pass
        self.original_stdin = original_stdin

        try:
            self.errors = (
                sys.stdin.errors
            )  # Who knew? sys streams have an errors attribute!
        except Exception:
            # Not sure if it's available in all Python versions...
            pass

    def readline(self, *args, **kwargs):
        return "\n"

    def read(self, *args, **kwargs):
        return self.readline()

    def write(self, *args, **kwargs):
        pass

    def flush(self, *args, **kwargs):
        pass

    def close(self, *args, **kwargs):
        pass


def binary_stdio():
    stdin, stdout = sys.stdin.buffer, sys.stdout.buffer

    # The original stdin cannot be used and anything written to stdout will
    # be put in the stderr.
    sys.stdin, sys.stdout = (_DummyStdin(), sys.stderr)

    return stdin, stdout


def socket_connect(host, port):
    import socket as socket_module

    s = socket_module.socket(socket_module.AF_INET, socket_module.SOCK_STREAM)

    #  Set TCP keepalive on an open socket.
    #  It activates after 1 second (TCP_KEEPIDLE,) of idleness,
    #  then sends a keepalive ping once every 3 seconds (TCP_KEEPINTVL),
    #  and closes the connection after 5 failed ping (TCP_KEEPCNT), or 15 seconds
    try:
        s.setsockopt(socket_module.SOL_SOCKET, socket_module.SO_KEEPALIVE, 1)
    except (AttributeError, OSError):
        pass  # May not be available everywhere.
    try:
        s.setsockopt(socket_module.IPPROTO_TCP, socket_module.TCP_KEEPIDLE, 1)
    except (AttributeError, OSError):
        pass  # May not be available everywhere.
    try:
        s.setsockopt(socket_module.IPPROTO_TCP, socket_module.TCP_KEEPINTVL, 3)
    except (AttributeError, OSError):
        pass  # May not be available everywhere.
    try:
        s.setsockopt(socket_module.IPPROTO_TCP, socket_module.TCP_KEEPCNT, 5)
    except (AttributeError, OSError):
        pass  # May not be available everywhere.

    try:
        # 10 seconds default timeout
        s.settimeout(DEFAULT_TIMEOUT if USE_TIMEOUTS else NO_TIMEOUT)
        s.connect((host, port))
        s.settimeout(None)  # no timeout after connected
    except Exception:
        raise RuntimeError("Could not connect to %s: %s", host, port)

    rfile = s.makefile("rb")
    wfile = s.makefile("wb")
    return rfile, wfile


def add_arguments(parser):
    parser.description = "Preload action-server actions"

    parser.add_argument(
        "--tcp", action="store_true", help="Use TCP server instead of stdio"
    )
    parser.add_argument("--host", default="127.0.0.1", help="Bind to this address")
    parser.add_argument("--port", default=-1, type=int, help="Bind to this port")


class MessagesHandler:
    def __init__(self, read_stream, write_stream):
        try:
            from preload_actions_streams import (  # type: ignore
                JsonRpcStreamReaderThread,
                JsonRpcStreamWriter,
            )
        except ImportError:
            from .preload_actions_streams import (
                JsonRpcStreamReaderThread,
                JsonRpcStreamWriter,
            )

        from queue import Queue

        self._readqueue = Queue()
        self._jsonrpc_stream_reader = JsonRpcStreamReaderThread(
            read_stream, self._readqueue, self._on_message
        )
        self._jsonrpc_stream_writer = JsonRpcStreamWriter(write_stream)

    def start(self):
        self._jsonrpc_stream_reader.start()

        # Removed for now: this messes up the logging because
        # by the time the imports are done it's important that the
        # needed scaffolding to collect the log contents is in place.
        #
        # import io
        # from contextlib import redirect_stderr, redirect_stdout
        # Collect actions so that it's ready to go when requested.
        # s = io.StringIO()
        # try:
        #     from robocorp.actions import cli
        #
        #     with redirect_stdout(s), redirect_stderr(s):
        #         cli.main(["list"], exit=False)
        # except BaseException:
        #     print(s.getvalue())
        #     traceback.print_exc()

        while True:
            msg = self._readqueue.get()
            self._on_message(msg)

    def _on_message(self, message):
        # Original command line is something as:
        # python = get_python_exe_from_env(env)
        # cmdline: List[str] = [
        #     python,
        #     "-m",
        #     "robocorp.actions",
        #     "run",
        #     "--preload-module",
        #     "preload_actions",
        #     "-a",
        #     action.name,
        # ]
        #
        # cmdline.append(str(action.file))
        # cmdline.append(f"--json-input={input_json}")

        # Some things must be set in the environment for the run:
        #
        # env["ROBOT_ARTIFACTS"] = robot_artifacts
        # env["RC_ACTION_RESULT_LOCATION"] = result_json
        command = message.get("command")
        if command == "run_action":
            from robocorp.actions import cli

            returncode = 1
            try:
                action_name = message["action_name"]
                action_file = message["action_file"]
                input_json = message["input_json"]
                robot_artifacts = message["robot_artifacts"]
                result_json = message["result_json"]
                headers = message["headers"]
                cookies = message["cookies"]
                reuse_process = message["reuse_process"]

                os.environ["ROBOT_ARTIFACTS"] = robot_artifacts
                os.environ["RC_ACTION_RESULT_LOCATION"] = result_json

                if reuse_process:
                    # Setup is skipped (for callbacks which still haven't been
                    # executed)
                    os.environ["RC_TASKS_SKIP_SESSION_SETUP"] = "1"

                    # Teardown is skipped (for all callbacks).
                    os.environ["RC_TASKS_SKIP_SESSION_TEARDOWN"] = "1"
                else:
                    os.environ.pop("RC_TASKS_SKIP_SESSION_TEARDOWN", None)
                    os.environ.pop("RC_TASKS_SKIP_SESSION_SETUP", None)

                if headers:
                    for key, value in headers.items():
                        if key and value and key.upper() == "X_ACTION_TRACE":
                            os.environ[key.upper()] = value

                # The preloaded actions must be always in place.
                sys.modules.pop("preload_actions", None)
                args = [
                    "run",
                    "--preload-module",
                    "preload_actions",
                    "-a",
                    action_name,
                    action_file,
                    f"--json-input={input_json}",
                ]

                returncode = cli.main(
                    args,
                    exit=False,
                    **self._plugin_manager_kwargs(
                        {"request": {"headers": headers, "cookies": cookies}}
                    ),
                )
            except BaseException:
                traceback.print_exc()

            finally:
                self._jsonrpc_stream_writer.write({"returncode": returncode})

    def _plugin_manager_kwargs(self, managed_parameters) -> Dict[str, Any]:
        try:
            from robocorp.actions._managed_parameters import ManagedParameters
            from robocorp.actions._request import Request
            from robocorp.tasks._customization._extension_points import (
                EPManagedParameters,
            )
            from robocorp.tasks._customization._plugin_manager import PluginManager

        except ImportError:
            return {}

        # Ok, we're dealing with a newer version of robocorp.actions and
        # robocorp.tasks, so add the customization of parameters to
        # add the 'request' parameter.

        try:
            pm = getattr(self, "_pm")
        except AttributeError:
            pm = self._pm = PluginManager()
        pm.set_instance(
            EPManagedParameters,
            ManagedParameters(
                {"request": Request.model_validate(managed_parameters["request"])}
            ),
        )

        return {"plugin_manager": pm}


def main(args=None):
    original_args = args if args is not None else sys.argv[1:]

    parser = argparse.ArgumentParser()
    add_arguments(parser)

    args = parser.parse_args(args=original_args)

    if args.tcp:
        rfile, wfile = socket_connect(args.host, args.port)
    else:
        rfile, wfile = binary_stdio()

    server = MessagesHandler(rfile, wfile)
    server.start()


if __name__ == "__main__":
    try:
        main()
    except Exception:
        # Critical error (the logging may not be set up properly).
        traceback.print_exc()



# Based on: https://github.com/conda/conda/blob/main/conda/models/match_spec.py
"""
Original license:
BSD 3-Clause License

Copyright (c) 2012, Anaconda, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the copyright holder nor the names of its
      contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import os
import re
from logging import getLogger

from . import conda_url

log = getLogger(__name__)


class InvalidMatchSpec(RuntimeError):
    pass


def _parse_version_plus_build(v_plus_b):
    """This should reliably pull the build string out of a version + build string combo.
    Examples:
        >>> _parse_version_plus_build("=1.2.3 0")
        ('=1.2.3', '0')
        >>> _parse_version_plus_build("1.2.3=0")
        ('1.2.3', '0')
        >>> _parse_version_plus_build(">=1.0 , < 2.0 py34_0")
        ('>=1.0,<2.0', 'py34_0')
        >>> _parse_version_plus_build(">=1.0 , < 2.0 =py34_0")
        ('>=1.0,<2.0', 'py34_0')
        >>> _parse_version_plus_build("=1.2.3 ")
        ('=1.2.3', None)
        >>> _parse_version_plus_build(">1.8,<2|==1.7")
        ('>1.8,<2|==1.7', None)
        >>> _parse_version_plus_build("* openblas_0")
        ('*', 'openblas_0')
        >>> _parse_version_plus_build("* *")
        ('*', '*')
    """
    parts = re.search(
        r"((?:.+?)[^><!,|]?)(?:(?<![=!|,<>~])(?:[ =])([^-=,|<>~]+?))?$", v_plus_b
    )
    if parts:
        version, build = parts.groups()
        build = build and build.strip()
    else:
        version, build = v_plus_b, None
    return version and version.replace(" ", ""), build


def _parse_legacy_dist(dist_str):
    """
    Examples:
        >>> _parse_legacy_dist("_license-1.1-py27_1.tar.bz2")
        ('_license', '1.1', 'py27_1')
        >>> _parse_legacy_dist("_license-1.1-py27_1")
        ('_license', '1.1', 'py27_1')
    """
    from .conda_path import strip_pkg_extension

    dist_str, _ = strip_pkg_extension(dist_str)
    name, version, build = dist_str.rsplit("-", 2)
    return name, version, build


def _parse_channel(channel_val):
    from .conda_channel import Channel

    if not channel_val:
        return None, None
    chn = Channel(channel_val)
    channel_name = chn.name or chn.base_url
    return channel_name, chn.subdir


_PARSE_CACHE: dict = {}


def parse_spec_str(spec_str):
    cached_result = _PARSE_CACHE.get(spec_str)
    if cached_result:
        return cached_result

    import warnings

    from .conda_channel import Channel
    from .conda_path import expand, is_package_file, url_to_path
    from .conda_url import is_url

    original_spec_str = spec_str

    # pre-step for ugly backward compat
    if spec_str.endswith("@"):
        feature_name = spec_str[:-1]
        return {
            "name": "*",
            "track_features": (feature_name,),
        }

    # Step 1. strip '#' comment
    if "#" in spec_str:
        ndx = spec_str.index("#")
        spec_str, _ = spec_str[:ndx], spec_str[ndx:]
        spec_str.strip()

    # Step 1.b strip ' if ' anticipating future compatibility issues
    spec_split = spec_str.split(" if ", 1)
    if len(spec_split) > 1:
        log.debug("Ignoring conditional in spec %s", spec_str)
    spec_str = spec_split[0]

    # Step 2. done if spec_str is a tarball
    if is_package_file(spec_str):
        # treat as a normal url
        if not is_url(spec_str):
            spec_str = conda_url.unquote(conda_url.path_to_url(expand(spec_str)))

        channel = Channel(spec_str)
        if channel.subdir:
            name, version, build = _parse_legacy_dist(channel.package_filename)
            result = {
                "channel": channel.canonical_name,
                "subdir": channel.subdir,
                "name": name,
                "version": version,
                "build": build,
                "fn": channel.package_filename,
                "url": spec_str,
            }
        else:
            # url is not a channel
            if spec_str.startswith("file://"):
                # We must undo percent-encoding when generating fn.
                path_or_url = url_to_path(spec_str)
            else:
                path_or_url = spec_str

            return {
                "name": "*",
                "fn": os.path.basename(path_or_url),
                "url": spec_str,
            }
        return result

    # Step 3. strip off brackets portion
    brackets = {}
    m3 = re.match(r".*(?:(\[.*\]))", spec_str)
    if m3:
        brackets_str = m3.groups()[0]
        spec_str = spec_str.replace(brackets_str, "")
        brackets_str = brackets_str[1:-1]
        m3b = re.finditer(
            r'([a-zA-Z0-9_-]+?)=(["\']?)([^\'"]*?)(\2)(?:[, ]|$)', brackets_str
        )
        for match in m3b:
            key, _, value, _ = match.groups()
            if not key or not value:
                raise InvalidMatchSpec(
                    original_spec_str, "key-value mismatch in brackets"
                )
            brackets[key] = value

    # Step 4. strip off parens portion
    m4 = re.match(r".*(?:(\(.*\)))", spec_str)
    parens = {}
    if m4:
        parens_str = m4.groups()[0]
        spec_str = spec_str.replace(parens_str, "")
        parens_str = parens_str[1:-1]
        m4b = re.finditer(
            r'([a-zA-Z0-9_-]+?)=(["\']?)([^\'"]*?)(\2)(?:[, ]|$)', parens_str
        )
        for match in m4b:
            key, _, value, _ = match.groups()
            parens[key] = value
        if "optional" in parens_str:
            parens["optional"] = True

    # Step 5. strip off '::' channel and namespace
    m5 = spec_str.rsplit(":", 2)
    m5_len = len(m5)
    if m5_len == 3:
        channel_str, namespace, spec_str = m5
    elif m5_len == 2:
        namespace, spec_str = m5
        channel_str = None
    elif m5_len:
        spec_str = m5[0]
        channel_str, namespace = None, None
    else:
        raise NotImplementedError()
    channel, subdir = _parse_channel(channel_str)
    if "channel" in brackets:
        b_channel, b_subdir = _parse_channel(brackets.pop("channel"))
        if b_channel:
            channel = b_channel
        if b_subdir:
            subdir = b_subdir
    if "subdir" in brackets:
        subdir = brackets.pop("subdir")

    # Step 6. strip off package name from remaining version + build
    m3 = re.match(r"([^ =<>!~]+)?([><!=~ ].+)?", spec_str)
    if m3:
        name, spec_str = m3.groups()
        if name is None:
            raise InvalidMatchSpec(
                original_spec_str, "no package name found in '%s'" % spec_str
            )
    else:
        raise InvalidMatchSpec(original_spec_str, "no package name found")

    # Step 7. otherwise sort out version + build
    spec_str = spec_str and spec_str.strip()
    # This was an attempt to make MatchSpec('numpy-1.11.0-py27_0') work like we'd want. It's
    # not possible though because plenty of packages have names with more than one '-'.
    # if spec_str is None and name.count('-') >= 2:
    #     name, version, build = _parse_legacy_dist(name)
    if spec_str:
        if "[" in spec_str:
            raise InvalidMatchSpec(
                original_spec_str, "multiple brackets sections not allowed"
            )

        version, build = _parse_version_plus_build(spec_str)

        # Catch cases where version ends up as "==" and pass it through so existing error
        # handling code can treat it like cases where version ends up being "<=" or ">=".
        # This is necessary because the "Translation" code below mangles "==" into a empty
        # string, which results in an empty version field on "components." The set of fields
        # on components drives future logic which breaks on an empty string but will deal with
        # missing versions like "==", "<=", and ">=" "correctly."
        #
        # All of these "missing version" cases result from match specs like "numpy==",
        # "numpy<=", "numpy>=", "numpy= " (with trailing space). Existing code indicates
        # these should be treated as an error and an exception raised.
        # IMPORTANT: "numpy=" (no trailing space) is treated as valid.
        if version == "==" or version == "=":
            pass
        # Otherwise,
        # translate version '=1.2.3' to '1.2.3*'
        # is it a simple version starting with '='? i.e. '=1.2.3'
        elif version[0] == "=":
            test_str = version[1:]
            if version[:2] == "==" and build is None:
                version = version[2:]
            elif not any(c in test_str for c in "=,|"):
                if build is None and test_str[-1] != "*":
                    version = test_str + "*"
                else:
                    version = test_str
    else:
        version, build = None, None

    # Step 8. now compile components together
    components = {}
    components["name"] = name or "*"

    if channel is not None:
        components["channel"] = channel
    if subdir is not None:
        components["subdir"] = subdir
    if namespace is not None:
        # components['namespace'] = namespace
        pass
    if version is not None:
        components["version"] = version
    if build is not None:
        components["build"] = build

    # anything in brackets will now strictly override key as set in other area of spec str
    # EXCEPT FOR: name
    # If we let name in brackets override a name outside of brackets it is possible to write
    # MatchSpecs that appear to install one package but actually install a completely different one
    # e.g. tensorflow[name=* version=* md5=<hash of pytorch package> ] will APPEAR to install
    # tensorflow but actually install pytorch.
    if "name" in components and "name" in brackets:
        warnings.warn(
            f"'name' specified both inside ({brackets['name']}) and outside ({components['name']})"
            " of brackets. the value outside of brackets ({components['name']}) will be used."
        )
        del brackets["name"]
    components.update(brackets)
    components["_original_spec_str"] = original_spec_str
    _PARSE_CACHE[original_spec_str] = components
    return components



import logging
from functools import cache

from fastapi import FastAPI
from fastapi.exceptions import HTTPException, RequestValidationError
from fastapi.middleware.cors import CORSMiddleware

from . import _errors
from ._settings import get_settings

LOGGER = logging.getLogger(__name__)


class _CustomFastAPI(FastAPI):
    def openapi(self):
        """
        As we're no longer using custom models for the routing APIs (instead of
        pydantic or generated methods for the API, the body is directly queried
        and interpreted as json and its schema validated), FastAPI is no longer
        adding the error references, so, we override the FastAPI.openapi()
        method to provide the needed components.
        """
        if self.openapi_schema:
            return self.openapi_schema

        openapi = FastAPI.openapi(self)
        components = openapi.setdefault("components", {})
        schemas = components.setdefault("schemas", {})
        schemas["HTTPValidationError"] = {
            "properties": {
                "errors": {
                    "items": {"$ref": "#/components/schemas/ValidationError"},
                    "type": "array",
                    "title": "Errors",
                }
            },
            "type": "object",
            "title": "HTTPValidationError",
        }
        schemas["ValidationError"] = {
            "properties": {
                "loc": {
                    "items": {"anyOf": [{"type": "string"}, {"type": "integer"}]},
                    "type": "array",
                    "title": "Location",
                },
                "msg": {"type": "string", "title": "Message"},
                "type": {"type": "string", "title": "Error Type"},
            },
            "type": "object",
            "required": ["loc", "msg", "type"],
            "title": "ValidationError",
        }

        return openapi


@cache
def get_app():
    settings = get_settings()

    server = {"url": settings.server_url}
    app = _CustomFastAPI(title=settings.title, servers=[server])

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_methods=["*"],
        allow_headers=["*"],
        allow_credentials=True,
    )

    app.add_exception_handler(_errors.RequestError, _errors.request_error_handler)
    app.add_exception_handler(HTTPException, _errors.http_error_handler)
    app.add_exception_handler(RequestValidationError, _errors.http422_error_handler)
    app.add_exception_handler(Exception, _errors.http500_error_handler)

    return app



from dataclasses import dataclass
from typing import Dict, Iterator, Optional

from ._deps_protocols import _RangeTypedDict
from .conda_impl import conda_match_spec, conda_version


@dataclass
class CondaDepInfo:
    name: str  # The name of the dep (i.e.: python)
    value: str  # The full value of the dep (i.e.: python=3.7)
    version_spec: str  # The version SPEC of the dep (as seen by conda: '3.7.*' or '>3.2')
    dep_range: _RangeTypedDict
    error_msg: Optional[str] = None

    def get_dep_vspec(self) -> Optional[conda_version.VersionSpec]:
        try:
            vspec = conda_version.VersionSpec(self.version_spec)
        except Exception:
            return None
        return vspec


class CondaDeps:
    def __init__(self) -> None:
        self._deps: Dict[str, CondaDepInfo] = {}

    def add_dep(self, value: str, dep_range: _RangeTypedDict):
        """
        Args:
            value: This is the value found in the spec. Something as:
            'python==3.7'.
        """
        try:
            spec = conda_match_spec.parse_spec_str(value)

            # It may not have a version if it wasn't specified.
            version_spec = spec.get("version", "*")
            name = spec["name"]
        except Exception:
            pass
        else:
            self._deps[name] = CondaDepInfo(name, value, version_spec, dep_range)

    def get_dep_vspec(self, spec_name: str) -> Optional[conda_version.VersionSpec]:
        conda_dep_info = self._deps.get(spec_name)
        if conda_dep_info is None:
            return None
        return conda_dep_info.get_dep_vspec()

    def get_dep_range(self, spec_name: str) -> _RangeTypedDict:
        return self._deps[spec_name].dep_range

    def iter_deps_infos(self) -> Iterator[CondaDepInfo]:
        yield from self._deps.values()



import io
import logging
import os
import sys
import zipfile
from urllib.request import urlopen

TEMPLATE_URL = (
    "https://github.com/robocorp/template-action/archive/refs/heads/master.zip"
)

log = logging.getLogger(__name__)


def create_new_project(directory: str = "."):
    try:
        if not directory:
            directory = input("Name of the project: ")
            if not directory:
                raise RuntimeError("The name of the project was not given.")

        if directory != ".":
            if os.path.exists(directory) and os.listdir(directory):
                raise RuntimeError(
                    f"The folder: {directory} already exists and is not empty."
                )

        download_and_unzip(TEMPLATE_URL, directory)

        log.info("✅ Project created")
    except KeyboardInterrupt:
        log.debug("Operation cancelled")
    except Exception as e:
        log.critical(f"Error creating the project: {e}")


def download_and_unzip(url: str, directory: str = "."):
    """
    Downloads a zip file from the given URL to memory,
    extracts files one by one and makes .sh files executable.
    """

    created_parents = set()
    with urlopen(url) as response:
        with io.BytesIO(response.read()) as zip_buffer:
            with zipfile.ZipFile(zip_buffer, "r") as zip_ref:
                namelist = list(zip_ref.namelist())

                for path_in_zip in namelist:
                    parts = path_in_zip.split("/", 1)
                    if len(parts) < 2 or not parts[1]:
                        continue

                    # i.e.: Remove the root dir from the .zip
                    # (which is something as template-action-master)
                    write_to = os.path.join(directory, parts[1])

                    if path_in_zip.endswith("/"):
                        created_parents.add(write_to)
                        os.makedirs(write_to, exist_ok=True)
                        continue
                    else:
                        parent_dir = os.path.dirname(write_to)
                        if parent_dir and parent_dir not in created_parents:
                            created_parents.add(parent_dir)
                            os.makedirs(parent_dir, exist_ok=True)

                    with zip_ref.open(path_in_zip, "r") as zip_file:
                        data = zip_file.read()

                    with open(write_to, "wb") as out_file:
                        out_file.write(data)

                    # Make .sh files executable
                    if path_in_zip.endswith(".sh"):
                        if sys.platform != "win32":
                            os.chmod(write_to, 0o755)


# if __name__ == "__main__":
#     download_and_unzip(TEMPLATE_URL, "ra")



import itertools
import json
import logging
import socket as socket_module
import subprocess
import sys
import threading
from contextlib import contextmanager
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Set

from starlette.requests import Request
from termcolor import colored

from robocorp.action_server._models import Action, ActionPackage

from ._settings import Settings, is_frozen

log = logging.getLogger(__name__)

AF_INET, SOCK_STREAM, SHUT_WR, SOL_SOCKET, SO_REUSEADDR, IPPROTO_TCP, socket = (
    socket_module.AF_INET,
    socket_module.SOCK_STREAM,
    socket_module.SHUT_WR,
    socket_module.SOL_SOCKET,
    socket_module.SO_REUSEADDR,
    socket_module.IPPROTO_TCP,
    socket_module.socket,
)

if sys.platform == "win32":
    SO_EXCLUSIVEADDRUSE = socket_module.SO_EXCLUSIVEADDRUSE


class _Key(tuple):
    pass


def _create_server_socket(host: str, port: int):
    try:
        server = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)
        if sys.platform == "win32":
            server.setsockopt(SOL_SOCKET, SO_EXCLUSIVEADDRUSE, 1)
        else:
            server.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)

        server.bind((host, port))
        server.settimeout(None)
    except Exception:
        server.close()
        raise

    return server


def _connect_to_socket(host, port):
    """connects to a host/port"""
    s = socket(AF_INET, SOCK_STREAM)

    #  Set TCP keepalive on an open socket.
    #  It activates after 1 second (TCP_KEEPIDLE,) of idleness,
    #  then sends a keepalive ping once every 3 seconds (TCP_KEEPINTVL),
    #  and closes the connection after 5 failed ping (TCP_KEEPCNT), or 15 seconds
    try:
        s.setsockopt(SOL_SOCKET, socket_module.SO_KEEPALIVE, 1)
    except (AttributeError, OSError):
        pass  # May not be available everywhere.
    try:
        s.setsockopt(socket_module.IPPROTO_TCP, socket_module.TCP_KEEPIDLE, 1)
    except (AttributeError, OSError):
        pass  # May not be available everywhere.
    try:
        s.setsockopt(socket_module.IPPROTO_TCP, socket_module.TCP_KEEPINTVL, 3)
    except (AttributeError, OSError):
        pass  # May not be available everywhere.
    try:
        s.setsockopt(socket_module.IPPROTO_TCP, socket_module.TCP_KEEPCNT, 5)
    except (AttributeError, OSError):
        pass  # May not be available everywhere.

    timeout = 20
    s.settimeout(timeout)
    s.connect((host, port))
    s.settimeout(None)  # no timeout after connected
    return s


class ProcessHandle:
    def __init__(self, settings: Settings, action_package: ActionPackage):
        from queue import Queue

        from robocorp.action_server._preload_actions.preload_actions_streams import (
            JsonRpcStreamWriter,
        )
        from robocorp.action_server._robo_utils.callback import Callback
        from robocorp.action_server._robo_utils.run_in_thread import run_in_thread

        from ._actions_run_helpers import (
            _add_preload_actions_dir_to_env_pythonpath,
            get_action_package_cwd,
        )
        from ._preload_actions.preload_actions_streams import JsonRpcStreamReaderThread
        from ._robo_utils.process import build_python_launch_env

        self.key = _get_process_handle_key(settings, action_package)
        env = json.loads(action_package.env_json)
        _add_preload_actions_dir_to_env_pythonpath(env)
        env = build_python_launch_env(env)

        if settings.reuse_processes:
            # When reusing processes we don't want to dump threads if
            # the process doesn't exit!
            env["RC_DUMP_THREADS_AFTER_RUN"] = "0"

        if "PYTHON_EXE" in env:
            python_exe = env["PYTHON_EXE"]
        else:
            if is_frozen():
                log.critical(
                    f"Unable to create process for action package: {action_package} "
                    "(environment does not contain PYTHON_EXE)."
                )
                return

            python_exe = sys.executable

        # stdin/stdout is no longer an option because numpy gets halted
        # if stdin is being read while importing numpy.
        # https://github.com/numpy/numpy/issues/24290
        # https://github.com/robocorp/robocorp/issues/271
        use_tcp = True

        cwd = get_action_package_cwd(settings, action_package)
        subprocess_kwargs = dict(
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=cwd,
            env=env,
        )

        def _process_stream_reader(stderr_or_stdout):
            while True:
                line_bytes = stderr_or_stdout.readline()
                if not line_bytes:
                    break
                line_as_str = line_bytes.decode("utf-8", "replace")
                print(
                    colored(f"output (pid: {pid}): ", attrs=["dark"])
                    + f"{line_as_str.strip()}\n",
                    end="",
                )
                self._on_output(line_bytes)

        self._read_queue: "Queue[dict]" = Queue()

        if use_tcp:
            server_socket = _create_server_socket("127.0.0.1", 0)
            host, port = server_socket.getsockname()
            cmdline = [
                python_exe,
                "-m",
                "preload_actions_server_main",
                "--tcp",
                f"--host={host}",
                f"--port={port}",
            ]

            def accept_connection():
                server_socket.listen(1)
                sock, _addr = server_socket.accept()
                return sock

            connection_future = run_in_thread(accept_connection)

            self._process = subprocess.Popen(cmdline, **subprocess_kwargs)
            self._on_output = Callback()

            pid = self._process.pid

            stderr = self._process.stderr
            stdout = self._process.stdout

            t = threading.Thread(
                target=_process_stream_reader, args=(stderr,), daemon=True
            )
            t.name = f"Stderr reader (pid: {pid})"
            t.start()

            t = threading.Thread(
                target=_process_stream_reader, args=(stdout,), daemon=True
            )
            t.name = f"Stdout reader (pid: {pid})"
            t.start()

            try:
                s = connection_future.result(10)
            except Exception:
                log.exception(
                    "Process that runs action did not connect back in the available timeout."
                )
                raise
            read_from = s.makefile("rb")
            write_to = s.makefile("wb")

            self._writer = JsonRpcStreamWriter(write_to, sort_keys=True)
            self._reader = JsonRpcStreamReaderThread(
                read_from, self._read_queue, lambda *args, **kwargs: None
            )
        else:
            # Will start things using the stdin/stdout for communicating.
            cmdline = [
                python_exe,
                "-m",
                "preload_actions_server_main",
            ]
            subprocess_kwargs["stdin"] = subprocess.PIPE

            self._process = subprocess.Popen(cmdline, **subprocess_kwargs)
            self._on_output = Callback()

            pid = self._process.pid

            stderr = self._process.stderr
            t = threading.Thread(target=_process_stream_reader, args=(stderr,))
            t.name = f"Stderr reader (pid: {pid})"
            t.start()

            write_to = self._process.stdin
            read_from = self._process.stdout
            self._writer = JsonRpcStreamWriter(write_to, sort_keys=True)
            self._reader = JsonRpcStreamReaderThread(
                read_from, self._read_queue, lambda *args, **kwargs: None
            )
        self._reader.start()

    @property
    def pid(self):
        return self._process.pid

    def is_alive(self) -> bool:
        from ._robo_utils.process import is_process_alive

        if self._process.poll() is not None:
            return False

        return is_process_alive(self._process.pid)

    def kill(self) -> None:
        from ._robo_utils.process import kill_process_and_subprocesses

        if not self.is_alive():
            return

        log.debug("Subprocess kill [pid=%s]", self._process.pid)
        kill_process_and_subprocesses(self._process.pid)

    def _do_run_action(
        self,
        action: Action,
        input_json: Path,
        robot_artifacts: Path,
        result_json: Path,
        request: Request,
        reuse_process: bool,
    ) -> int:
        msg = {
            "command": "run_action",
            "action_name": action.name,
            "action_file": f"{action.file}",
            "input_json": f"{input_json}",
            "robot_artifacts": f"{robot_artifacts}",
            "result_json": f"{result_json}",
            "headers": dict(request.headers),
            "cookies": dict(request.cookies),
            "reuse_process": reuse_process,
        }
        self._writer.write(msg)
        queue = self._read_queue
        result_msg = queue.get(block=True)
        return result_msg["returncode"]

    def run_action(
        self,
        action: Action,
        input_json: Path,
        robot_artifacts: Path,
        output_file: Path,
        result_json: Path,
        request: Request,
        reuse_process: bool,
    ) -> int:
        """
        Runs the action and returns the returncode from running the action.

        (returncode=0 means everything is Ok).
        """
        with output_file.open("wb") as stream:

            def on_output(line_bytes: bytes):
                stream.write(line_bytes)

            with self._on_output.register(on_output):
                # stdout is now used for communicating, so, don't hear on it.
                returncode = self._do_run_action(
                    action,
                    input_json,
                    robot_artifacts,
                    result_json,
                    request,
                    reuse_process,
                )
                return returncode


def _get_process_handle_key(settings: Settings, action_package: ActionPackage) -> _Key:
    """
    Given an action provides a key where the key identifies whether a
    given action can be run at a given ProcessHandle.
    """
    from ._actions_run_helpers import get_action_package_cwd

    env = tuple(sorted(json.loads(action_package.env_json).items()))
    cwd = get_action_package_cwd(settings, action_package)
    return _Key((env, cwd))


class ActionsProcessPool:
    def __init__(
        self,
        settings: Settings,
        action_package_id_to_action_package: Dict[str, ActionPackage],
        actions: List[Action],
    ):
        self._settings = settings
        self.action_package_id_to_action_package = action_package_id_to_action_package

        # We just want the actions which are enabled.
        self.actions = [action for action in actions if action.enabled]

        # An iterator which keeps cycling over the actions.
        self._cycle_actions_iterator = itertools.cycle(self.actions)

        self._lock = threading.Lock()
        self._running_processes: Dict[_Key, Set[ProcessHandle]] = {}
        self._idle_processes: Dict[_Key, Set[ProcessHandle]] = {}

        # Semaphore used to track running processes.
        self._processes_running_semaphore = threading.Semaphore(self.max_processes)

        self._warmup_processes()

    @property
    def _reuse_processes(self) -> bool:
        """
        Returns:
            Whether processes can be reused.
        """
        return self._settings.reuse_processes

    @property
    def max_processes(self) -> int:
        """
        Returns:
            The maximum number of processes that may be created by the process
            pool.
        """
        return self._settings.max_processes

    @property
    def min_processes(self) -> int:
        return self._settings.min_processes

    def _create_process(self, action: Action):
        action_package: ActionPackage = self.action_package_id_to_action_package[
            action.action_package_id
        ]

        process_handle = ProcessHandle(self._settings, action_package)
        assert self._lock.locked(), "Lock must be acquired at this point."
        self._add_to_idle_processes(process_handle)

    def dispose(self):
        with self._lock:
            for processes in itertools.chain(
                self._idle_processes.values(), self._running_processes.values()
            ):
                for process_handle in processes:
                    process_handle.kill()
            self._idle_processes.clear()
            self._running_processes.clear()

    def get_idle_processes_count(self) -> int:
        with self._lock:
            return self._get_idle_processes_count_unlocked()

    def _get_idle_processes_count_unlocked(self) -> int:
        assert self._lock.locked(), "Lock must be acquired at this point."
        count = 0
        for v in self._idle_processes.values():
            count += len(v)
        return count

    def get_running_processes_count(self) -> int:
        with self._lock:
            return self._get_running_processes_count_unlocked()

    def _get_running_processes_count_unlocked(self) -> int:
        assert self._lock.locked(), "Lock must be acquired at this point."
        count = 0
        for v in self._running_processes.values():
            count += len(v)
        return count

    def _count_total_processes(self) -> int:
        assert self._lock.locked(), "Lock must be acquired at this point."
        count = 0
        for v in itertools.chain(
            self._running_processes.values(), self._idle_processes.values()
        ):
            count += len(v)
        return count

    def _add_to_idle_processes(self, process_handle: ProcessHandle):
        assert self._lock.locked(), "Lock must be acquired at this point."
        processes = self._idle_processes.get(process_handle.key)
        if processes is None:
            processes = self._idle_processes[process_handle.key] = set()
        processes.add(process_handle)

    def _add_to_running_processes(self, process_handle: ProcessHandle):
        assert self._lock.locked(), "Lock must be acquired at this point."
        self._processes_running_semaphore.acquire()
        processes = self._running_processes.get(process_handle.key)
        if processes is None:
            processes = self._running_processes[process_handle.key] = set()
        processes.add(process_handle)

    def _remove_from_running_processes(self, process_handle: ProcessHandle):
        assert self._lock.locked(), "Lock must be acquired at this point."
        self._processes_running_semaphore.release()
        processes = self._running_processes.get(process_handle.key)
        if not processes:
            return
        processes.discard(process_handle)

    def _warmup_processes(self):
        if not self.actions:
            return

        with self._lock:
            while self._count_total_processes() < self._settings.min_processes:
                one_action = next(self._cycle_actions_iterator)
                self._create_process(one_action)

    @contextmanager
    def obtain_process_for_action(self, action: Action) -> Iterator[ProcessHandle]:
        action_package: ActionPackage = self.action_package_id_to_action_package[
            action.action_package_id
        ]

        key = _get_process_handle_key(self._settings, action_package)
        process_handle: Optional[ProcessHandle] = None
        while True:
            with self._lock:
                processes = self._idle_processes.get(key)
                if processes:
                    # Get any process from the (compatible) idle processes.
                    process_handle = processes.pop()
                    log.debug(
                        f"Process Pool: Using idle process ({process_handle.pid})."
                    )
                    if not process_handle.is_alive():
                        # Process died while trying to get it.
                        log.critical(
                            f"Process Pool: Unexpected: Idle process exited "
                            f"({process_handle.pid})."
                        )
                        continue

                    self._add_to_running_processes(process_handle)
                else:
                    # No compatible process: we need to create one now.
                    n_running = self._get_running_processes_count_unlocked()
                    if n_running < self.max_processes:
                        self._create_process(action)
                        processes = self._idle_processes.get(key)
                        assert (
                            processes
                        ), f"Expected idle processes bound to key: {key} at this point!"
                        process_handle = processes.pop()
                        log.debug(
                            f"Process Pool: Created process ({process_handle.pid})."
                        )
                        if not process_handle.is_alive():
                            # Process died while trying to get it.
                            log.critical(
                                f"Process Pool: Unexpected: Idle process exited right "
                                f"after creation ({process_handle.pid})."
                            )
                            continue
                        self._add_to_running_processes(process_handle)
                    else:
                        log.critical(
                            f"Delayed running action: {action.name} because "
                            f"{self.max_processes} actions are already running ("
                            f"waiting for another action to finish running)."
                        )

            if process_handle is not None:
                break
            else:
                # Each 5 seconds check again and print delayed message if still
                # not able to run.
                if self._processes_running_semaphore.acquire(timeout=5):
                    self._processes_running_semaphore.release()

        if process_handle is not None:
            try:
                yield process_handle
            finally:
                with self._lock:
                    self._remove_from_running_processes(process_handle)
                    if process_handle.is_alive():
                        if self._reuse_processes:
                            curr_idle = self._get_idle_processes_count_unlocked()
                            if self.min_processes <= curr_idle:
                                log.debug(
                                    f"Process Pool: Exited process ({process_handle.pid}) -- min processes already satisfied."
                                )
                                # We cannot reuse it!
                                process_handle.kill()
                            else:
                                log.debug(
                                    f"Process Pool: Adding back to pool ({process_handle.pid})."
                                )
                                self._add_to_idle_processes(process_handle)
                        else:
                            log.debug(
                                f"Process Pool: Exited process ({process_handle.pid}) -- not reusing processes."
                            )
                            # We cannot reuse it!
                            process_handle.kill()

                # If needed recreate idle processes which were removed (needed
                # especially when not reusing processes, but if some process
                # crashes it's also needed).
                self._warmup_processes()

            return

        raise AssertionError("Expected process_handle to be not None!")


_actions_process_pool: Optional[ActionsProcessPool] = None


@contextmanager
def setup_actions_process_pool(
    settings: Settings,
    action_package_id_to_action_package: Dict[str, ActionPackage],
    actions: List[Action],
):
    global _actions_process_pool

    _actions_process_pool = ActionsProcessPool(
        settings, action_package_id_to_action_package, actions
    )
    yield
    _actions_process_pool = None


def get_actions_process_pool() -> ActionsProcessPool:
    assert _actions_process_pool is not None
    return _actions_process_pool



if __name__ == "__main__":
    from robocorp.action_server.cli import main

    args = None
    # args = "start -v -p 8090".split()
    # args = ["start"]
    # import os
    # os.chdir(r"X:\temp\check-action-server\action_package")
    # args = ["start", "--expose"]

    main(args)
